@@ -48,38 +48,43 @@ bool SourceTextModuleDescriptor::ModuleRequestComparer::operator()(
             rhs->import_attributes()->size());
   }
 
+  if (lhs->phase() != rhs->phase()) {
+    return lhs->phase() < rhs->phase();
+  }
+
   return false;
 }
 
 void SourceTextModuleDescriptor::AddImport(
     const AstRawString* import_name, const AstRawString* local_name,
-    const AstRawString* module_request,
+    const AstRawString* specifier, const ModuleImportPhase import_phase,
     const ImportAttributes* import_attributes, const Scanner::Location loc,
     const Scanner::Location specifier_loc, Zone* zone) {
   Entry* entry = zone->New<Entry>(loc);
   entry->local_name = local_name;
   entry->import_name = import_name;
-  entry->module_request =
-      AddModuleRequest(module_request, import_attributes, specifier_loc, zone);
+  entry->module_request = AddModuleRequest(
+      specifier, import_phase, import_attributes, specifier_loc, zone);
   AddRegularImport(entry);
 }
 
 void SourceTextModuleDescriptor::AddStarImport(
-    const AstRawString* local_name, const AstRawString* module_request,
+    const AstRawString* local_name, const AstRawString* specifier,
     const ImportAttributes* import_attributes, const Scanner::Location loc,
     const Scanner::Location specifier_loc, Zone* zone) {
   Entry* entry = zone->New<Entry>(loc);
   entry->local_name = local_name;
   entry->module_request =
-      AddModuleRequest(module_request, import_attributes, specifier_loc, zone);
+      AddModuleRequest(specifier, ModuleImportPhase::kEvaluation,
+                       import_attributes, specifier_loc, zone);
   AddNamespaceImport(entry, zone);
 }
 
 void SourceTextModuleDescriptor::AddEmptyImport(
-    const AstRawString* module_request,
-    const ImportAttributes* import_attributes,
+    const AstRawString* specifier, const ImportAttributes* import_attributes,
     const Scanner::Location specifier_loc, Zone* zone) {
-  AddModuleRequest(module_request, import_attributes, specifier_loc, zone);
+  AddModuleRequest(specifier, ModuleImportPhase::kEvaluation, import_attributes,
+                   specifier_loc, zone);
 }
 
 void SourceTextModuleDescriptor::AddExport(const AstRawString* local_name,
@@ -93,44 +98,47 @@ void SourceTextModuleDescriptor::AddExport(const AstRawString* local_name,
 
 void SourceTextModuleDescriptor::AddExport(
     const AstRawString* import_name, const AstRawString* export_name,
-    const AstRawString* module_request,
-    const ImportAttributes* import_attributes, const Scanner::Location loc,
-    const Scanner::Location specifier_loc, Zone* zone) {
+    const AstRawString* specifier, const ImportAttributes* import_attributes,
+    const Scanner::Location loc, const Scanner::Location specifier_loc,
+    Zone* zone) {
   DCHECK_NOT_NULL(import_name);
   DCHECK_NOT_NULL(export_name);
   Entry* entry = zone->New<Entry>(loc);
   entry->export_name = export_name;
   entry->import_name = import_name;
   entry->module_request =
-      AddModuleRequest(module_request, import_attributes, specifier_loc, zone);
+      AddModuleRequest(specifier, ModuleImportPhase::kEvaluation,
+                       import_attributes, specifier_loc, zone);
   AddSpecialExport(entry, zone);
 }
 
 void SourceTextModuleDescriptor::AddStarExport(
-    const AstRawString* module_request,
-    const ImportAttributes* import_attributes, const Scanner::Location loc,
-    const Scanner::Location specifier_loc, Zone* zone) {
+    const AstRawString* specifier, const ImportAttributes* import_attributes,
+    const Scanner::Location loc, const Scanner::Location specifier_loc,
+    Zone* zone) {
   Entry* entry = zone->New<Entry>(loc);
   entry->module_request =
-      AddModuleRequest(module_request, import_attributes, specifier_loc, zone);
+      AddModuleRequest(specifier, ModuleImportPhase::kEvaluation,
+                       import_attributes, specifier_loc, zone);
   AddSpecialExport(entry, zone);
 }
 
 namespace {
 template <typename IsolateT>
-Handle<PrimitiveHeapObject> ToStringOrUndefined(IsolateT* isolate,
-                                                const AstRawString* s) {
+Handle<UnionOf<String, Undefined>> ToStringOrUndefined(IsolateT* isolate,
+                                                       const AstRawString* s) {
   if (s == nullptr) return isolate->factory()->undefined_value();
   return s->string();
 }
 }  // namespace
 
 template <typename IsolateT>
-Handle<ModuleRequest> SourceTextModuleDescriptor::AstModuleRequest::Serialize(
+DirectHandle<ModuleRequest>
+SourceTextModuleDescriptor::AstModuleRequest::Serialize(
     IsolateT* isolate) const {
   // The import attributes will be stored in this array in the form:
   // [key1, value1, location1, key2, value2, location2, ...]
-  Handle<FixedArray> import_attributes_array =
+  DirectHandle<FixedArray> import_attributes_array =
       isolate->factory()->NewFixedArray(
           static_cast<int>(import_attributes()->size() *
                            ModuleRequest::kAttributeEntrySize),
@@ -149,17 +157,18 @@ Handle<ModuleRequest> SourceTextModuleDescriptor::AstModuleRequest::Serialize(
     }
   }
   return v8::internal::ModuleRequest::New(isolate, specifier()->string(),
-                                          import_attributes_array, position());
+                                          phase_, import_attributes_array,
+                                          position());
 }
-template Handle<ModuleRequest>
+template DirectHandle<ModuleRequest>
 SourceTextModuleDescriptor::AstModuleRequest::Serialize(Isolate* isolate) const;
-template Handle<ModuleRequest>
+template DirectHandle<ModuleRequest>
 SourceTextModuleDescriptor::AstModuleRequest::Serialize(
     LocalIsolate* isolate) const;
 
 template <typename IsolateT>
-Handle<SourceTextModuleInfoEntry> SourceTextModuleDescriptor::Entry::Serialize(
-    IsolateT* isolate) const {
+DirectHandle<SourceTextModuleInfoEntry>
+SourceTextModuleDescriptor::Entry::Serialize(IsolateT* isolate) const {
   CHECK(Smi::IsValid(module_request));  // TODO(neis): Check earlier?
   return SourceTextModuleInfoEntry::New(
       isolate, ToStringOrUndefined(isolate, export_name),
@@ -167,19 +176,19 @@ Handle<SourceTextModuleInfoEntry> SourceTextModuleDescriptor::Entry::Serialize(
       ToStringOrUndefined(isolate, import_name), module_request, cell_index,
       location.beg_pos, location.end_pos);
 }
-template Handle<SourceTextModuleInfoEntry>
+template DirectHandle<SourceTextModuleInfoEntry>
 SourceTextModuleDescriptor::Entry::Serialize(Isolate* isolate) const;
-template Handle<SourceTextModuleInfoEntry>
+template DirectHandle<SourceTextModuleInfoEntry>
 SourceTextModuleDescriptor::Entry::Serialize(LocalIsolate* isolate) const;
 
 template <typename IsolateT>
-Handle<FixedArray> SourceTextModuleDescriptor::SerializeRegularExports(
+DirectHandle<FixedArray> SourceTextModuleDescriptor::SerializeRegularExports(
     IsolateT* isolate, Zone* zone) const {
   // We serialize regular exports in a way that lets us later iterate over their
   // local names and for each local name immediately access all its export
   // names.  (Regular exports have neither import name nor module request.)
 
-  ZoneVector<Handle<Object>> data(
+  ZoneVector<IndirectHandle<Object>> data(
       SourceTextModuleInfo::kRegularExportLength * regular_exports_.size(),
       zone);
   int index = 0;
@@ -220,17 +229,19 @@ Handle<FixedArray> SourceTextModuleDescriptor::SerializeRegularExports(
 
   // We cannot create the FixedArray earlier because we only now know the
   // precise size.
-  Handle<FixedArray> result =
+  DirectHandle<FixedArray> result =
       isolate->factory()->NewFixedArray(index, AllocationType::kOld);
   for (int i = 0; i < index; ++i) {
     result->set(i, *data[i]);
   }
   return result;
 }
-template Handle<FixedArray> SourceTextModuleDescriptor::SerializeRegularExports(
-    Isolate* isolate, Zone* zone) const;
-template Handle<FixedArray> SourceTextModuleDescriptor::SerializeRegularExports(
-    LocalIsolate* isolate, Zone* zone) const;
+template DirectHandle<FixedArray>
+SourceTextModuleDescriptor::SerializeRegularExports(Isolate* isolate,
+                                                    Zone* zone) const;
+template DirectHandle<FixedArray>
+SourceTextModuleDescriptor::SerializeRegularExports(LocalIsolate* isolate,
+                                                    Zone* zone) const;
 
 void SourceTextModuleDescriptor::MakeIndirectExportsExplicit(Zone* zone) {
   for (auto it = regular_exports_.begin(); it != regular_exports_.end();) {
