@@ -9,7 +9,7 @@
 #include "src/compiler/graph-zone-traits.h"
 #include "src/compiler/opcodes.h"
 #include "src/compiler/operator.h"
-#include "src/compiler/types.h"
+#include "src/compiler/turbofan-types.h"
 #include "src/zone/zone-containers.h"
 
 namespace v8 {
@@ -18,8 +18,7 @@ namespace compiler {
 
 // Forward declarations.
 class Edge;
-class Graph;
-
+class TFGraph;
 
 // Marks are used during traversal of the graph to distinguish states of nodes.
 // Each node has a mark which is a monotonically increasing integer, and a
@@ -418,7 +417,7 @@ class V8_EXPORT_PRIVATE Node::Inputs final {
   int count_;
 };
 
-// An encapsulation for information associated with a single use of node as a
+// An encapsulation for information associated with a single use of a node as an
 // input from another node, allowing access to both the defining node and
 // the node having the input.
 class Edge final {
