@@ -69,7 +69,7 @@ class Interpreter {
   void SetBytecodeHandler(Bytecode bytecode, OperandScale operand_scale,
                           Tagged<Code> handler);
 
-  V8_EXPORT_PRIVATE Handle<JSObject> GetDispatchCountersObject();
+  V8_EXPORT_PRIVATE DirectHandle<JSObject> GetDispatchCountersObject();
 
   void ForEachBytecode(const std::function<void(Bytecode, OperandScale)>& f);
 
