@@ -27,6 +27,7 @@ class SafepointEntry;
 class RootVisitor;
 
 enum class Builtin;
+enum class LazyDeoptimizeReason : uint8_t;
 
 // Code is a container for data fields related to its associated
 // {InstructionStream} object. Since {InstructionStream} objects reside on
@@ -52,6 +53,7 @@ enum class Builtin;
 //  |           ...            |  <-- MS + handler_table_offset()
 //  |                          |  <-- MS + constant_pool_offset()
 //  |                          |  <-- MS + code_comments_offset()
+//  |                          |  <-- MS + builtin_jump_table_info_offset()
 //  |                          |  <-- MS + unwinding_info_offset()
 //  +--------------------------+  <-- MetadataEnd()
 //
@@ -101,8 +103,8 @@ class Code : public ExposedTrustedObject {
   inline void UpdateInstructionStart(IsolateForSandbox isolate,
                                      Tagged<InstructionStream> istream);
 
-  inline void initialize_flags(CodeKind kind, bool is_turbofanned,
-                               int stack_slots);
+  inline void initialize_flags(CodeKind kind, bool is_context_specialized,
+                               bool is_turbofanned);
 
   // Clear uninitialized padding space. This ensures that the snapshot content
   // is deterministic.
@@ -113,17 +115,28 @@ class Code : public ExposedTrustedObject {
   void FlushICache() const;
 
   DECL_PRIMITIVE_ACCESSORS(can_have_weak_objects, bool)
-  DECL_PRIMITIVE_ACCESSORS(marked_for_deoptimization, bool)
+  DECL_PRIMITIVE_GETTER(marked_for_deoptimization, bool)
 
   DECL_PRIMITIVE_ACCESSORS(metadata_size, int)
   // [handler_table_offset]: The offset where the exception handler table
   // starts.
   DECL_PRIMITIVE_ACCESSORS(handler_table_offset, int)
+  // [builtin_jump_table_info offset]: Offset of the builtin jump table info.
+  DECL_PRIMITIVE_ACCESSORS(builtin_jump_table_info_offset, int32_t)
   // [unwinding_info_offset]: Offset of the unwinding info section.
   DECL_PRIMITIVE_ACCESSORS(unwinding_info_offset, int32_t)
   // [deoptimization_data]: Array containing data for deopt for non-baseline
   // code.
-  DECL_ACCESSORS(deoptimization_data, Tagged<TrustedFixedArray>)
+  DECL_ACCESSORS(deoptimization_data, Tagged<ProtectedFixedArray>)
+  // [parameter_count]: The number of formal parameters, including the
+  // receiver. Currently only available for optimized functions.
+  // TODO(saelo): make this always available. This is just a matter of figuring
+  // out how to obtain the parameter count during code generation when no
+  // BytecodeArray is available from which it can be copied.
+  DECL_PRIMITIVE_ACCESSORS(parameter_count, uint16_t)
+  inline uint16_t parameter_count_without_receiver() const;
+  DECL_PRIMITIVE_ACCESSORS(wasm_js_tagged_parameter_count, uint16_t)
+  DECL_PRIMITIVE_ACCESSORS(wasm_js_first_tagged_parameter, uint16_t)
 
   // Whether this type of Code uses deoptimization data, in which case the
   // deoptimization_data field will be populated.
@@ -167,7 +180,7 @@ class Code : public ExposedTrustedObject {
   DECL_ACCESSORS(wrapper, Tagged<CodeWrapper>)
 
   // Unchecked accessors to be used during GC.
-  inline Tagged<TrustedFixedArray> unchecked_deoptimization_data() const;
+  inline Tagged<ProtectedFixedArray> unchecked_deoptimization_data() const;
 
   DECL_RELAXED_UINT32_ACCESSORS(flags)
 
@@ -186,6 +199,7 @@ class Code : public ExposedTrustedObject {
 
   // Tells whether the code checks the tiering state in the function's feedback
   // vector.
+  // TODO(olivfi, 42204201): Remove this once leaptiering is enabled everywhere.
   inline bool checks_tiering_state() const;
 
   // Tells whether the outgoing parameters of this code are tagged pointers.
@@ -199,6 +213,10 @@ class Code : public ExposedTrustedObject {
   // TurboFan optimizing compiler.
   inline bool is_turbofanned() const;
 
+  // [is_context_specialized]: Tells whether the code object was specialized to
+  // a constant context.
+  inline bool is_context_specialized() const;
+
   // [uses_safepoint_table]: Whether this InstructionStream object uses
   // safepoint tables (note the table may still be empty, see
   // has_safepoint_table).
@@ -206,10 +224,12 @@ class Code : public ExposedTrustedObject {
 
   // [stack_slots]: If {uses_safepoint_table()}, the number of stack slots
   // reserved in the code prologue; otherwise 0.
-  inline int stack_slots() const;
+  inline uint32_t stack_slots() const;
 
   inline Tagged<TrustedByteArray> SourcePositionTable(
       Isolate* isolate, Tagged<SharedFunctionInfo> sfi) const;
+  int SourcePosition(int offset) const;
+  int SourceStatementPosition(int offset) const;
 
   inline Address safepoint_table_address() const;
   inline int safepoint_table_size() const;
@@ -227,6 +247,10 @@ class Code : public ExposedTrustedObject {
   inline int code_comments_size() const;
   inline bool has_code_comments() const;
 
+  inline Address builtin_jump_table_info() const;
+  inline int builtin_jump_table_info_size() const;
+  inline bool has_builtin_jump_table_info() const;
+
   inline Address unwinding_info_start() const;
   inline Address unwinding_info_end() const;
   inline int unwinding_info_size() const;
@@ -245,6 +269,11 @@ class Code : public ExposedTrustedObject {
   inline Address metadata_start() const;
   inline Address metadata_end() const;
 
+#ifdef V8_ENABLE_LEAPTIERING
+  inline void set_js_dispatch_handle(JSDispatchHandle handle);
+  inline JSDispatchHandle js_dispatch_handle() const;
+#endif  // V8_ENABLE_LEAPTIERING
+
   // The size of the associated InstructionStream object, if it exists.
   inline int InstructionStreamObjectSize() const;
 
@@ -275,7 +304,10 @@ class Code : public ExposedTrustedObject {
   SafepointEntry GetSafepointEntry(Isolate* isolate, Address pc);
   MaglevSafepointEntry GetMaglevSafepointEntry(Isolate* isolate, Address pc);
 
-  void SetMarkedForDeoptimization(Isolate* isolate, const char* reason);
+  inline void SetMarkedForDeoptimization(Isolate* isolate,
+                                         LazyDeoptimizeReason reason);
+  void TraceMarkForDeoptimization(Isolate* isolate,
+                                  LazyDeoptimizeReason reason);
 
   inline bool CanContainWeakObjects();
   inline bool IsWeakObject(Tagged<HeapObject> object);
@@ -284,7 +316,7 @@ class Code : public ExposedTrustedObject {
       Tagged<Object> object);
 
   // This function should be called only from GC.
-  void ClearEmbeddedObjects(Heap* heap);
+  void ClearEmbeddedObjectsAndJSDispatchHandles(Heap* heap);
 
   // [embedded_objects_cleared]: If CodeKindIsOptimizedJSFunction(kind), tells
   // whether the embedded objects in the code marked for deoptimization were
@@ -334,53 +366,60 @@ class Code : public ExposedTrustedObject {
                  Address current_pc = kNullAddress);
 #endif
 
-  DECL_CAST(Code)
   DECL_VERIFIER(Code)
 
 // Layout description.
-#define CODE_DATA_FIELDS(V)                                                   \
-  /* The deoptimization_data_or_interpreter_data field contains: */           \
-  /*  - A DeoptimizationData for optimized code (maglev or turbofan) */       \
-  /*  - A BytecodeArray or InterpreterData for baseline code */               \
-  /*  - Smi::zero() for all other types of code (e.g. builtin) */             \
-  V(kDeoptimizationDataOrInterpreterDataOffset, kTaggedSize)                  \
-  /* This field contains: */                                                  \
-  /*  - A bytecode offset table (trusted byte array) for baseline code */     \
-  /*  - A (possibly empty) source position table (trusted byte array) for */  \
-  /*    most other types of code */                                           \
-  /*  - Smi::zero() for embedded builtin code (in RO space) */                \
-  /*    TODO(saelo) once we have a  trusted RO space, we could instead use */ \
-  /*    empty_trusted_byte_array to avoid using Smi::zero() at all. */        \
-  V(kPositionTableOffset, kTaggedSize)                                        \
-  /* Strong pointer fields. */                                                \
-  V(kStartOfStrongFieldsOffset, 0)                                            \
-  V(kWrapperOffset, kTaggedSize)                                              \
-  V(kEndOfStrongFieldsWithMainCageBaseOffset, 0)                              \
-  /* The InstructionStream field is special: it uses code_cage_base. */       \
-  V(kInstructionStreamOffset, kTaggedSize)                                    \
-  V(kEndOfStrongFieldsOffset, 0)                                              \
-  /* Untagged data not directly visited by GC starts here. */                 \
-  /* When the sandbox is off, the instruction_start field contains a raw */   \
-  /* pointer to the first instruction of this Code. */                        \
-  /* If the sandbox is on, this field does not exist. Instead, the */         \
-  /* instruction_start is stored in this Code's code pointer table entry */   \
-  /* referenced via the kSelfIndirectPointerOffset field */                   \
-  V(kInstructionStartOffset, V8_ENABLE_SANDBOX_BOOL ? 0 : kSystemPointerSize) \
-  /* The serializer needs to copy bytes starting from here verbatim. */       \
-  V(kFlagsOffset, kUInt32Size)                                                \
-  V(kInstructionSizeOffset, kIntSize)                                         \
-  V(kMetadataSizeOffset, kIntSize)                                            \
-  /* TODO(jgruber): TF-specific fields could be merged with builtin_id. */    \
-  V(kInlinedBytecodeSizeOffset, kIntSize)                                     \
-  V(kOsrOffsetOffset, kInt32Size)                                             \
-  V(kHandlerTableOffsetOffset, kIntSize)                                      \
-  V(kUnwindingInfoOffsetOffset, kInt32Size)                                   \
-  V(kConstantPoolOffsetOffset, V8_EMBEDDED_CONSTANT_POOL_BOOL ? kIntSize : 0) \
-  V(kCodeCommentsOffsetOffset, kIntSize)                                      \
-  /* TODO(jgruber): 12 bits would suffice, steal from here if needed. */      \
-  V(kBuiltinIdOffset, kInt16Size)                                             \
-  V(kUnalignedSize, OBJECT_POINTER_PADDING(kUnalignedSize))                   \
-  /* Total size. */                                                           \
+#define CODE_DATA_FIELDS(V)                                                    \
+  /* The deoptimization_data_or_interpreter_data field contains: */            \
+  /*  - A DeoptimizationData for optimized code (maglev or turbofan) */        \
+  /*  - A BytecodeArray or InterpreterData for baseline code */                \
+  /*  - Smi::zero() for all other types of code (e.g. builtin) */              \
+  V(kDeoptimizationDataOrInterpreterDataOffset, kTaggedSize)                   \
+  /* This field contains: */                                                   \
+  /*  - A bytecode offset table (trusted byte array) for baseline code */      \
+  /*  - A (possibly empty) source position table (trusted byte array) for */   \
+  /*    most other types of code */                                            \
+  /*  - Smi::zero() for embedded builtin code (in RO space) */                 \
+  /*    TODO(saelo) once we have a  trusted RO space, we could instead use */  \
+  /*    empty_trusted_byte_array to avoid using Smi::zero() at all. */         \
+  V(kPositionTableOffset, kTaggedSize)                                         \
+  /* Strong pointer fields. */                                                 \
+  V(kStartOfStrongFieldsOffset, 0)                                             \
+  V(kWrapperOffset, kTaggedSize)                                               \
+  V(kEndOfStrongFieldsWithMainCageBaseOffset, 0)                               \
+  /* The InstructionStream field is special: it uses code_cage_base. */        \
+  V(kInstructionStreamOffset, kTaggedSize)                                     \
+  V(kEndOfStrongFieldsOffset, 0)                                               \
+  /* Untagged data not directly visited by GC starts here. */                  \
+  /* When the sandbox is off, the instruction_start field contains a raw */    \
+  /* pointer to the first instruction of this Code. */                         \
+  /* If the sandbox is on, this field does not exist. Instead, the */          \
+  /* instruction_start is stored in this Code's code pointer table entry */    \
+  /* referenced via the kSelfIndirectPointerOffset field */                    \
+  V(kInstructionStartOffset, V8_ENABLE_SANDBOX_BOOL ? 0 : kSystemPointerSize)  \
+  /* The serializer needs to copy bytes starting from here verbatim. */        \
+  V(kDispatchHandleOffset,                                                     \
+    V8_ENABLE_LEAPTIERING_BOOL ? kJSDispatchHandleSize : 0)                    \
+  V(kFlagsOffset, kUInt32Size)                                                 \
+  V(kInstructionSizeOffset, kIntSize)                                          \
+  V(kMetadataSizeOffset, kIntSize)                                             \
+  /* TODO(jgruber): TF-specific fields could be merged with builtin_id. */     \
+  V(kInlinedBytecodeSizeOffset, kIntSize)                                      \
+  V(kOsrOffsetOffset, kInt32Size)                                              \
+  V(kHandlerTableOffsetOffset, kIntSize)                                       \
+  V(kUnwindingInfoOffsetOffset, kInt32Size)                                    \
+  V(kConstantPoolOffsetOffset, V8_EMBEDDED_CONSTANT_POOL_BOOL ? kIntSize : 0)  \
+  V(kCodeCommentsOffsetOffset, kIntSize)                                       \
+  V(kBuiltinJumpTableInfoOffsetOffset,                                         \
+    V8_BUILTIN_JUMP_TABLE_INFO_BOOL ? kInt32Size : 0)                          \
+  /* This field is currently only used during deoptimization. If this space */ \
+  /* is ever needed for other purposes, it would probably be possible to */    \
+  /* obtain the parameter count from the BytecodeArray instead. */             \
+  V(kParameterCountOffset, kUInt16Size)                                        \
+  /* TODO(jgruber): 12 bits would suffice, steal from here if needed. */       \
+  V(kBuiltinIdOffset, kInt16Size)                                              \
+  V(kUnalignedSize, OBJECT_POINTER_PADDING(kUnalignedSize))                    \
+  /* Total size. */                                                            \
   V(kSize, 0)
 
   DEFINE_FIELD_OFFSET_CONSTANTS(ExposedTrustedObject::kHeaderSize,
@@ -403,14 +442,12 @@ class Code : public ExposedTrustedObject {
 #define FLAGS_BIT_FIELDS(V, _)                \
   V(KindField, CodeKind, 4, _)                \
   V(IsTurbofannedField, bool, 1, _)           \
-  /* Steal bits from here if needed: */       \
-  V(StackSlotsField, int, 24, _)              \
+  V(IsContextSpecializedField, bool, 1, _)    \
   V(MarkedForDeoptimizationField, bool, 1, _) \
   V(EmbeddedObjectsClearedField, bool, 1, _)  \
   V(CanHaveWeakObjectsField, bool, 1, _)
   DEFINE_BIT_FIELDS(FLAGS_BIT_FIELDS)
 #undef FLAGS_BIT_FIELDS
-  static_assert(FLAGS_BIT_FIELDS_Ranges::kBitsCount == 32);
   static_assert(FLAGS_BIT_FIELDS_Ranges::kBitsCount <=
                 FIELD_SIZE(kFlagsOffset) * kBitsPerByte);
   static_assert(kCodeKindCount <= KindField::kNumValues);
@@ -420,11 +457,14 @@ class Code : public ExposedTrustedObject {
       MarkedForDeoptimizationField::kShift;
   static const int kIsTurbofannedBit = IsTurbofannedField::kShift;
 
-  // Reserve one argument count value as the "don't adapt arguments" sentinel.
   static const int kArgumentsBits = 16;
-  static const int kMaxArguments = (1 << kArgumentsBits) - 2;
+  // Slightly less than 2^kArgumentBits-1 to allow for extra implicit arguments
+  // on the call nodes without overflowing the uint16_t input_count.
+  static const int kMaxArguments = (1 << kArgumentsBits) - 10;
 
  private:
+  DECL_PRIMITIVE_SETTER(marked_for_deoptimization, bool)
+
   inline void set_instruction_start(IsolateForSandbox isolate, Address value);
 
   // TODO(jgruber): These field names are incomplete, we've squashed in more
@@ -472,8 +512,6 @@ class Code : public ExposedTrustedObject {
 // (checked) casts do not work on GcSafeCode.
 class GcSafeCode : public HeapObject {
  public:
-  DECL_CAST(GcSafeCode)
-
   // Use with care, this casts away knowledge that we're dealing with a
   // special-semantics object.
   inline Tagged<Code> UnsafeCastToCode() const;
@@ -495,7 +533,7 @@ class GcSafeCode : public HeapObject {
   inline Tagged<Object> raw_instruction_stream() const;
   inline Address constant_pool() const;
   inline Address safepoint_table_address() const;
-  inline int stack_slots() const;
+  inline uint32_t stack_slots() const;
 
   inline int GetOffsetFromInstructionStart(Isolate* isolate, Address pc) const;
   inline Address InstructionStart(Isolate* isolate, Address pc) const;
@@ -503,6 +541,10 @@ class GcSafeCode : public HeapObject {
   inline bool CanDeoptAt(Isolate* isolate, Address pc) const;
   inline Tagged<Object> raw_instruction_stream(
       PtrComprCageBase code_cage_base) const;
+  // The two following accessors repurpose the InlinedBytecodeSize field, see
+  // comment in code-inl.h.
+  inline uint16_t wasm_js_tagged_parameter_count() const;
+  inline uint16_t wasm_js_first_tagged_parameter() const;
 
  private:
   OBJECT_CONSTRUCTORS(GcSafeCode, HeapObject);
@@ -515,7 +557,6 @@ class CodeWrapper : public Struct {
  public:
   DECL_CODE_POINTER_ACCESSORS(code)
 
-  DECL_CAST(CodeWrapper)
   DECL_PRINTER(CodeWrapper)
   DECL_VERIFIER(CodeWrapper)
 
