@@ -5,8 +5,10 @@
 #ifndef V8_OBJECTS_JS_ARRAY_BUFFER_INL_H_
 #define V8_OBJECTS_JS_ARRAY_BUFFER_INL_H_
 
-#include "src/heap/heap-write-barrier-inl.h"
 #include "src/objects/js-array-buffer.h"
+// Include the non-inl header before the rest of the headers.
+
+#include "src/heap/heap-write-barrier-inl.h"
 #include "src/objects/js-objects-inl.h"
 #include "src/objects/objects-inl.h"
 
@@ -30,6 +32,12 @@ RELEASE_ACQUIRE_ACCESSORS(JSTypedArray, base_pointer, Tagged<Object>,
                           kBasePointerOffset)
 
 size_t JSArrayBuffer::byte_length() const {
+  // Use GetByteLength() for growable SharedArrayBuffers.
+  DCHECK(!is_shared() || !is_resizable_by_js());
+  return byte_length_unchecked();
+}
+
+size_t JSArrayBuffer::byte_length_unchecked() const {
   return ReadBoundedSizeField(kRawByteLengthOffset);
 }
 
@@ -63,8 +71,9 @@ std::shared_ptr<BackingStore> JSArrayBuffer::GetBackingStore() const {
 size_t JSArrayBuffer::GetByteLength() const {
   if (V8_UNLIKELY(is_shared() && is_resizable_by_js())) {
     // Invariant: byte_length for GSAB is 0 (it needs to be read from the
-    // BackingStore).
-    DCHECK_EQ(0, byte_length());
+    // BackingStore). Don't use the byte_length getter, which DCHECKs that it's
+    // not used on growable SharedArrayBuffers.
+    DCHECK_EQ(0, byte_length_unchecked());
 
     // If the byte length is read after the JSArrayBuffer object is allocated
     // but before it's attached to the backing store, GetBackingStore returns
@@ -88,6 +97,16 @@ void JSArrayBuffer::SetBackingStoreRefForSerialization(uint32_t ref) {
   WriteField<Address>(kBackingStoreOffset, static_cast<Address>(ref));
 }
 
+void JSArrayBuffer::init_extension() {
+#if V8_COMPRESS_POINTERS
+  // The extension field is lazily-initialized, so set it to null initially.
+  base::AsAtomic32::Release_Store(extension_handle_location(),
+                                  kNullExternalPointerHandle);
+#else
+  base::AsAtomicPointer::Release_Store(extension_location(), nullptr);
+#endif  // V8_COMPRESS_POINTERS
+}
+
 ArrayBufferExtension* JSArrayBuffer::extension() const {
 #if V8_COMPRESS_POINTERS
   // We need Acquire semantics here when loading the entry, see below.
@@ -105,32 +124,29 @@ ArrayBufferExtension* JSArrayBuffer::extension() const {
 
 void JSArrayBuffer::set_extension(ArrayBufferExtension* extension) {
 #if V8_COMPRESS_POINTERS
-  if (extension != nullptr) {
-    Isolate* isolate = GetIsolateFromWritableObject(*this);
-    ExternalPointerTable& table = isolate->external_pointer_table();
-
-    // The external pointer handle for the extension is initialized lazily and
-    // so has to be zero here since, once set, the extension field can only be
-    // cleared, but not changed.
-    DCHECK_EQ(0, base::AsAtomic32::Relaxed_Load(extension_handle_location()));
+  // TODO(saelo): if we ever use the external pointer table for all external
+  // pointer fields in the no-sandbox-ptr-compression config, replace this code
+  // here and above with the respective external pointer accessors.
+  IsolateForPointerCompression isolate = GetIsolateFromWritableObject(*this);
+  const ExternalPointerTag tag = kArrayBufferExtensionTag;
+  Address value = reinterpret_cast<Address>(extension);
+  ExternalPointerTable& table = isolate.GetExternalPointerTableFor(tag);
 
+  ExternalPointerHandle current_handle =
+      base::AsAtomic32::Relaxed_Load(extension_handle_location());
+  if (current_handle == kNullExternalPointerHandle) {
     // We need Release semantics here, see above.
     ExternalPointerHandle handle = table.AllocateAndInitializeEntry(
-        isolate->heap()->external_pointer_space(),
-        reinterpret_cast<Address>(extension), kArrayBufferExtensionTag);
+        isolate.GetExternalPointerTableSpaceFor(tag, address()), value, tag);
     base::AsAtomic32::Release_Store(extension_handle_location(), handle);
+    EXTERNAL_POINTER_WRITE_BARRIER(*this, kExtensionOffset, tag);
   } else {
-    // This special handling of nullptr is required as it is used to initialize
-    // the slot, but is also beneficial when an ArrayBuffer is detached as it
-    // allows the external pointer table entry to be reclaimed while the
-    // ArrayBuffer is still alive.
-    base::AsAtomic32::Release_Store(extension_handle_location(),
-                                    kNullExternalPointerHandle);
+    table.Set(current_handle, value, tag);
   }
 #else
   base::AsAtomicPointer::Release_Store(extension_location(), extension);
 #endif  // V8_COMPRESS_POINTERS
-  WriteBarrier::Marking(*this, extension);
+  WriteBarrier::ForArrayBufferExtension(*this, extension);
 }
 
 #if V8_COMPRESS_POINTERS
@@ -199,7 +215,7 @@ void JSArrayBufferView::set_byte_length(size_t value) {
 }
 
 bool JSArrayBufferView::WasDetached() const {
-  return JSArrayBuffer::cast(buffer())->was_detached();
+  return Cast<JSArrayBuffer>(buffer())->was_detached();
 }
 
 BIT_FIELD_ACCESSORS(JSArrayBufferView, bit_field, is_length_tracking,
@@ -361,27 +377,26 @@ bool JSTypedArray::is_on_heap(AcquireLoadTag tag) const {
 }
 
 // static
-MaybeHandle<JSTypedArray> JSTypedArray::Validate(Isolate* isolate,
-                                                 Handle<Object> receiver,
-                                                 const char* method_name) {
+MaybeDirectHandle<JSTypedArray> JSTypedArray::Validate(
+    Isolate* isolate, DirectHandle<Object> receiver, const char* method_name) {
   if (V8_UNLIKELY(!IsJSTypedArray(*receiver))) {
     const MessageTemplate message = MessageTemplate::kNotTypedArray;
-    THROW_NEW_ERROR(isolate, NewTypeError(message), JSTypedArray);
+    THROW_NEW_ERROR(isolate, NewTypeError(message));
   }
 
-  Handle<JSTypedArray> array = Handle<JSTypedArray>::cast(receiver);
+  DirectHandle<JSTypedArray> array = Cast<JSTypedArray>(receiver);
   if (V8_UNLIKELY(array->WasDetached())) {
     const MessageTemplate message = MessageTemplate::kDetachedOperation;
-    Handle<String> operation =
+    DirectHandle<String> operation =
         isolate->factory()->NewStringFromAsciiChecked(method_name);
-    THROW_NEW_ERROR(isolate, NewTypeError(message, operation), JSTypedArray);
+    THROW_NEW_ERROR(isolate, NewTypeError(message, operation));
   }
 
   if (V8_UNLIKELY(array->IsVariableLength() && array->IsOutOfBounds())) {
     const MessageTemplate message = MessageTemplate::kDetachedOperation;
-    Handle<String> operation =
+    DirectHandle<String> operation =
         isolate->factory()->NewStringFromAsciiChecked(method_name);
-    THROW_NEW_ERROR(isolate, NewTypeError(message, operation), JSTypedArray);
+    THROW_NEW_ERROR(isolate, NewTypeError(message, operation));
   }
 
   // spec describes to return `buffer`, but it may disrupt current
