@@ -5,17 +5,19 @@
 #ifndef V8_REGEXP_REGEXP_AST_H_
 #define V8_REGEXP_REGEXP_AST_H_
 
+#include <optional>
+
 #include "src/base/strings.h"
 #include "src/regexp/regexp-flags.h"
 #include "src/zone/zone-containers.h"
 #include "src/zone/zone-list.h"
 #include "src/zone/zone.h"
+
 #ifdef V8_INTL_SUPPORT
 #include "unicode/uniset.h"
 #endif  // V8_INTL_SUPPORT
 
-namespace v8 {
-namespace internal {
+namespace v8::internal {
 
 #define FOR_EACH_REG_EXP_TREE_TYPE(VISIT) \
   VISIT(Disjunction)                      \
@@ -213,7 +215,6 @@ class RegExpTree : public ZoneObject {
 #undef MAKE_ASTYPE
 };
 
-
 class RegExpDisjunction final : public RegExpTree {
  public:
   explicit RegExpDisjunction(ZoneList<RegExpTree*>* alternatives);
@@ -236,7 +237,6 @@ class RegExpDisjunction final : public RegExpTree {
   int max_match_;
 };
 
-
 class RegExpAlternative final : public RegExpTree {
  public:
   explicit RegExpAlternative(ZoneList<RegExpTree*>* nodes);
@@ -256,7 +256,6 @@ class RegExpAlternative final : public RegExpTree {
   int max_match_;
 };
 
-
 class RegExpAssertion final : public RegExpTree {
  public:
   enum class Type {
@@ -300,7 +299,7 @@ class CharacterSet final {
 
  private:
   ZoneList<CharacterRange>* ranges_ = nullptr;
-  base::Optional<StandardCharacterSet> standard_set_type_;
+  std::optional<StandardCharacterSet> standard_set_type_;
 };
 
 class RegExpClassRanges final : public RegExpTree {
@@ -543,15 +542,16 @@ class RegExpText final : public RegExpTree {
   int length_ = 0;
 };
 
-
 class RegExpQuantifier final : public RegExpTree {
  public:
   enum QuantifierType { GREEDY, NON_GREEDY, POSSESSIVE };
-  RegExpQuantifier(int min, int max, QuantifierType type, RegExpTree* body)
+  RegExpQuantifier(int min, int max, QuantifierType type, int index,
+                   RegExpTree* body)
       : body_(body),
         min_(min),
         max_(max),
-        quantifier_type_(type) {
+        quantifier_type_(type),
+        index_(index) {
     if (min > 0 && body->min_match() > kInfinity / min) {
       min_match_ = kInfinity;
     } else {
@@ -575,6 +575,7 @@ class RegExpQuantifier final : public RegExpTree {
   int min() const { return min_; }
   int max() const { return max_; }
   QuantifierType quantifier_type() const { return quantifier_type_; }
+  int index() const { return index_; }
   bool is_possessive() const { return quantifier_type_ == POSSESSIVE; }
   bool is_non_greedy() const { return quantifier_type_ == NON_GREEDY; }
   bool is_greedy() const { return quantifier_type_ == GREEDY; }
@@ -587,9 +588,9 @@ class RegExpQuantifier final : public RegExpTree {
   int min_match_;
   int max_match_;
   QuantifierType quantifier_type_;
+  int index_;
 };
 
-
 class RegExpCapture final : public RegExpTree {
  public:
   explicit RegExpCapture(int index)
@@ -704,7 +705,6 @@ class RegExpLookaround final : public RegExpTree {
   int index_;
 };
 
-
 class RegExpBackReference final : public RegExpTree {
  public:
   explicit RegExpBackReference(Zone* zone) : captures_(1, zone) {}
@@ -731,7 +731,6 @@ class RegExpBackReference final : public RegExpTree {
   const ZoneVector<base::uc16>* name_ = nullptr;
 };
 
-
 class RegExpEmpty final : public RegExpTree {
  public:
   DECL_BOILERPLATE(Empty);
@@ -739,8 +738,7 @@ class RegExpEmpty final : public RegExpTree {
   int max_match() override { return 0; }
 };
 
-}  // namespace internal
-}  // namespace v8
+}  // namespace v8::internal
 
 #undef DECL_BOILERPLATE
 
