@@ -14,7 +14,7 @@ import {
 import { MovableView } from "./movable-view";
 import { SelectionBroker } from "../selection/selection-broker";
 import { SelectionMap } from "../selection/selection-map";
-import { TurboshaftGraphNode } from "../phases/turboshaft-graph-phase/turboshaft-graph-node";
+import { TurboshaftGraphOperation } from "../phases/turboshaft-graph-phase/turboshaft-graph-operation";
 import { TurboshaftGraphEdge } from "../phases/turboshaft-graph-phase/turboshaft-graph-edge";
 import { TurboshaftGraph } from "../turboshaft-graph";
 import { TurboshaftGraphLayout } from "../turboshaft-graph-layout";
@@ -40,7 +40,7 @@ export class TurboshaftGraphView extends MovableView<TurboshaftGraph> {
   graphLayout: TurboshaftGraphLayout;
   blockSelectionHandler: BlockSelectionHandler & ClearableHandler;
   visibleBlocks: d3.Selection<any, TurboshaftGraphBlock, any, any>;
-  visibleNodes: d3.Selection<any, TurboshaftGraphNode, any, any>;
+  visibleNodes: d3.Selection<any, TurboshaftGraphOperation, any, any>;
   visibleEdges: d3.Selection<any, TurboshaftGraphEdge<TurboshaftGraphBlock>, any, any>;
   visibleBubbles: d3.Selection<any, any, any, any>;
   blockDrag: d3.DragBehavior<any, TurboshaftGraphBlock, TurboshaftGraphBlock>;
@@ -96,6 +96,9 @@ export class TurboshaftGraphView extends MovableView<TurboshaftGraph> {
     this.phaseName = data.name;
     this.addCustomDataSelect(data.customData);
 
+    this.addToggleImgInput("toggle-compact-view",  "compact view",
+      this.state.compactView, partial(this.toggleCompactViewAction, this));
+
     const adaptedSelection = this.createGraph(data, rememberedSelection);
     this.broker.addNodeHandler(this.nodeSelectionHandler);
     this.broker.addBlockHandler(this.blockSelectionHandler);
@@ -198,7 +201,7 @@ export class TurboshaftGraphView extends MovableView<TurboshaftGraph> {
       if (query.length == 0) return;
 
       const reg = new RegExp(query);
-      const filterFunction = (node: TurboshaftGraphNode) => {
+      const filterFunction = (node: TurboshaftGraphOperation) => {
         if (!onlyVisible) node.block.collapsed = false;
         const customDataTitle = this.graph.customData.getTitle(node.id, DataTarget.Nodes);
         return (!onlyVisible || !node.block.collapsed) &&
@@ -262,7 +265,7 @@ export class TurboshaftGraphView extends MovableView<TurboshaftGraph> {
   private initializeNodeSelectionHandler(): NodeSelectionHandler & ClearableHandler {
     const view = this;
     return {
-      select: function (selectedNodes: Array<TurboshaftGraphNode>, selected: boolean,
+      select: function (selectedNodes: Array<TurboshaftGraphOperation>, selected: boolean,
                         scrollIntoView: boolean) {
         const sourcePositions = new Array<SourcePosition>();
         const nodes = new Set<string>();
@@ -402,7 +405,7 @@ export class TurboshaftGraphView extends MovableView<TurboshaftGraph> {
     this.visibleEdges
       .selectAll<SVGPathElement, TurboshaftGraphEdge<TurboshaftGraphBlock>>("path")
       .filter(edge => edge.target === block || edge.source === block)
-      .attr("d", edge => edge.generatePath(this.graph, this.nodesCustomDataShowed()));
+      .attr("d", edge => edge.generatePath(this.graph, this.nodesCustomDataShowed(), this.state.compactView));
   }
 
   private updateVisibleBlocksAndEdges(): void {
@@ -484,7 +487,7 @@ export class TurboshaftGraphView extends MovableView<TurboshaftGraph> {
       .attr("rx", C.TURBOSHAFT_BLOCK_BORDER_RADIUS)
       .attr("ry", C.TURBOSHAFT_BLOCK_BORDER_RADIUS)
       .attr("width", block => block.getWidth())
-      .attr("height", block => block.getHeight(view.nodesCustomDataShowed()));
+      .attr("height", block => block.getHeight(view.nodesCustomDataShowed(), this.state?.compactView));
 
     newBlocks.each(function (block: TurboshaftGraphBlock) {
       const svg = d3.select<SVGGElement, TurboshaftGraphBlock>(this);
@@ -529,7 +532,7 @@ export class TurboshaftGraphView extends MovableView<TurboshaftGraph> {
       .classed("selected", block => view.state.blocksSelection.isSelected(block))
       .attr("transform", block => `translate(${block.x},${block.y})`)
       .select("rect")
-      .attr("height", block =>  block.getHeight(view.nodesCustomDataShowed()));
+      .attr("height", block =>  block.getHeight(view.nodesCustomDataShowed(), this.state?.compactView));
 
     newAndOldBlocks.select("image")
       .attr("xlink:href", block => `${iconsPath}collapse_${block.collapsed ? "down" : "up"}.svg`);
@@ -537,7 +540,7 @@ export class TurboshaftGraphView extends MovableView<TurboshaftGraph> {
     newAndOldBlocks.select(".block-collapsed-label")
       .attr("visibility", block => block.collapsed ? "visible" : "hidden");
 
-    newAndOldEdges.attr("d", edge => edge.generatePath(view.graph, view.nodesCustomDataShowed()));
+    newAndOldEdges.attr("d", edge => edge.generatePath(view.graph, view.nodesCustomDataShowed(), view.state.compactView));
   }
 
   private appendInlineNodes(svg: d3.Selection<SVGGElement, TurboshaftGraphBlock, any, any>,
@@ -545,7 +548,7 @@ export class TurboshaftGraphView extends MovableView<TurboshaftGraph> {
     const state = this.state;
     const graph = this.graph;
     const filteredNodes = [...block.nodes.filter(_ => graph.isRendered())];
-    const allNodes = svg.selectAll<SVGGElement, TurboshaftGraphNode>(".inline-node");
+    const allNodes = svg.selectAll<SVGGElement, TurboshaftGraphOperation>(".inline-node");
     const selNodes = allNodes.data(filteredNodes, node => node.toString());
 
     // remove old nodes
@@ -563,7 +566,7 @@ export class TurboshaftGraphView extends MovableView<TurboshaftGraph> {
     const customData = this.graph.customData;
     const storageKey = this.customDataStorageKey();
     const selectedCustomData = storageGetItem(storageKey, null, false);
-    newNodes.each(function (node: TurboshaftGraphNode) {
+    newNodes.each(function (node: TurboshaftGraphOperation) {
       const nodeSvg = d3.select(this);
       nodeSvg
         .attr("id", node.id)
@@ -572,35 +575,35 @@ export class TurboshaftGraphView extends MovableView<TurboshaftGraph> {
         .classed("inline-node-label", true)
         .attr("dy", nodeY)
         .append("tspan")
-        .text(node.displayLabel)
+        .html(node.buildOperationText(state.compactView))
         .append("title")
         .text(`${node.getTitle()}${customData.getTitle(node.id, DataTarget.Nodes)}`);
 
       nodeSvg
-        .on("mouseenter", (node: TurboshaftGraphNode) => {
-          view.visibleNodes.data<TurboshaftGraphNode>(
+        .on("mouseenter", (node: TurboshaftGraphOperation) => {
+          view.visibleNodes.data<TurboshaftGraphOperation>(
             node.inputs.map(edge => edge.source), source => source.toString())
             .classed("input", true);
-          view.visibleNodes.data<TurboshaftGraphNode>(
+          view.visibleNodes.data<TurboshaftGraphOperation>(
             node.outputs.map(edge => edge.target), target => target.toString())
             .classed("output", true);
           view.hoveredNodeIdentifier = node.identifier();
           view.updateGraphVisibility();
         })
-        .on("mouseleave", (node: TurboshaftGraphNode) => {
+        .on("mouseleave", (node: TurboshaftGraphOperation) => {
           const inOutNodes = node.inputs.map(edge => edge.source)
             .concat(node.outputs.map(edge => edge.target));
-          view.visibleNodes.data<TurboshaftGraphNode>(inOutNodes, inOut => inOut.toString())
+          view.visibleNodes.data<TurboshaftGraphOperation>(inOutNodes, inOut => inOut.toString())
             .classed("input output", false);
           view.hoveredNodeIdentifier = null;
           view.updateGraphVisibility();
         })
-        .on("click", (node: TurboshaftGraphNode) => {
+        .on("click", (node: TurboshaftGraphOperation) => {
           if (!d3.event.shiftKey) view.nodeSelectionHandler.clear();
           view.nodeSelectionHandler.select([node], undefined, false);
           d3.event.stopPropagation();
         });
-      nodeY += node.labelBox.height;
+      nodeY += node.getHeight(false, state.compactView);
       if (view.graph.customData.nodes.size > 0) {
         const customData = view.graph.getCustomData(selectedCustomData, node.id, DataTarget.Nodes);
         nodeSvg
@@ -612,7 +615,7 @@ export class TurboshaftGraphView extends MovableView<TurboshaftGraph> {
           .text(view.getReadableString(customData, blockWidth))
           .append("title")
           .text(customData);
-        nodeY += node.labelBox.height;
+          nodeY += node.labelBox.height;
       }
     });
 
@@ -624,26 +627,37 @@ export class TurboshaftGraphView extends MovableView<TurboshaftGraph> {
     const view = this;
     const state = this.state;
     const showCustomData = this.nodesCustomDataShowed();
+    const customData = this.graph.customData;
+ 
     let totalHeight = 0;
     let blockId = 0;
-    view.visibleNodes.each(function (node: TurboshaftGraphNode) {
+    view.visibleNodes.each(function (node: TurboshaftGraphOperation) {
       const nodeSvg = d3.select(this);
       if (blockId != node.block.id) {
         blockId = node.block.id;
         totalHeight = 0;
       }
-      totalHeight += node.getHeight(showCustomData);
+      totalHeight += node.getHeight(showCustomData, state?.compactView);
       const nodeY = showCustomData ? totalHeight - node.labelBox.height : totalHeight;
 
       nodeSvg
         .select(".inline-node-label")
         .classed("selected", node => state.selection.isSelected(node))
         .attr("dy", nodeY)
-        .attr("visibility", !node.block.collapsed ? "visible" : "hidden");
+        .attr("visibility", !node.block.collapsed ? "visible" : "hidden")
+        .select("tspan")
+        .html(node.buildOperationText(state.compactView))
+        .append("title")
+        .text(`${node.getTitle()}${customData.getTitle(node.id, DataTarget.Nodes)}`);
 
+      // node.getHeight() == 0 iff node is fully inlined in compact view, so we
+      // don't want this to be visible.
+      const isVisible = !node.block.collapsed && showCustomData &&
+        node.getHeight(showCustomData, state.compactView) > 0;
       nodeSvg
         .select(".inline-node-custom-data")
-        .attr("visibility", !node.block.collapsed && showCustomData ? "visible" : "hidden");
+        .attr("dy", nodeY + node.labelBox.height)
+        .attr("visibility", isVisible ? "visible" : "hidden");
     });
   }
 
@@ -652,7 +666,7 @@ export class TurboshaftGraphView extends MovableView<TurboshaftGraph> {
     const storageKey = this.customDataStorageKey();
     const selectedCustomData = storageGetItem(storageKey, null, false);
     if (!this.nodesCustomDataShowed()) return;
-    view.visibleNodes.each(function (node: TurboshaftGraphNode) {
+    view.visibleNodes.each(function (node: TurboshaftGraphOperation) {
       const customData = view.graph.getCustomData(selectedCustomData, node.id, DataTarget.Nodes);
       d3.select(this)
         .select(".inline-node-custom-data")
@@ -677,7 +691,7 @@ export class TurboshaftGraphView extends MovableView<TurboshaftGraph> {
     }
     if (block.outputs.length > 0) {
       const x = block.getOutputX();
-      const y = block.getHeight(this.nodesCustomDataShowed()) + C.DEFAULT_NODE_BUBBLE_RADIUS;
+      const y = block.getHeight(this.nodesCustomDataShowed(), this.state?.compactView) + C.DEFAULT_NODE_BUBBLE_RADIUS;
       svg.append("circle")
         .classed("filledBubbleStyle", true)
         .attr("id", `ob,${block.id}`)
@@ -693,7 +707,7 @@ export class TurboshaftGraphView extends MovableView<TurboshaftGraph> {
       if (components[0] === "ob") {
         const from = view.graph.blockMap[components[1]];
         const x = from.getOutputX();
-        const y = from.getHeight(view.nodesCustomDataShowed()) + C.DEFAULT_NODE_BUBBLE_RADIUS;
+        const y = from.getHeight(view.nodesCustomDataShowed(), view.state?.compactView) + C.DEFAULT_NODE_BUBBLE_RADIUS;
         this.setAttribute("transform", `translate(${x},${y})`);
       }
     });
@@ -715,8 +729,8 @@ export class TurboshaftGraphView extends MovableView<TurboshaftGraph> {
           maxX = maxX ? Math.max(maxX, block.x + block.getWidth()) : block.x + block.getWidth();
           minY = minY ? Math.min(minY, block.y) : block.y;
           maxY = maxY
-            ? Math.max(maxY, block.y + block.getHeight(this.nodesCustomDataShowed()))
-            : block.y + block.getHeight(this.nodesCustomDataShowed());
+            ? Math.max(maxY, block.y + block.getHeight(this.nodesCustomDataShowed(), this.state?.compactView))
+            : block.y + block.getHeight(this.nodesCustomDataShowed(), this.state?.compactView);
         }
         if (blockHasSelection) {
           hasSelection = true;
@@ -817,6 +831,17 @@ export class TurboshaftGraphView extends MovableView<TurboshaftGraph> {
     view.focusOnSvg();
   }
 
+  private toggleCompactViewAction(view: TurboshaftGraphView): void {
+    view.state.compactView = !view.state.compactView;
+
+    const element = document.getElementById("toggle-compact-view");
+    element.classList.toggle("button-input-toggled", view.state.compactView);
+    const extent = view.graph.redetermineGraphBoundingBox(view.state.showCustomData);
+    view.panZoom.translateExtent(extent);
+    view.adaptiveUpdateGraphVisibility();
+    view.updateInlineNodesCustomData();
+  }
+
   private toggleCustomDataAction(view: TurboshaftGraphView): void {
     view.state.showCustomData = !view.state.showCustomData;
     const ranksMaxBlockHeight = new Array<number>();
@@ -825,7 +850,7 @@ export class TurboshaftGraphView extends MovableView<TurboshaftGraph> {
       ranksMaxBlockHeight[block.rank] = Math.max(ranksMaxBlockHeight[block.rank] ?? 0,
         block.collapsed
           ? block.height
-          : block.getHeight(view.nodesCustomDataShowed()));
+          : block.getHeight(view.nodesCustomDataShowed(), this.state?.compactView));
     }
 
     for (const block of view.graph.blocks()) {
@@ -854,7 +879,7 @@ export class TurboshaftGraphView extends MovableView<TurboshaftGraph> {
     this.updateGraphVisibility();
   }
 
-  private collapseUnusedBlocks(usedNodes: Iterable<TurboshaftGraphNode>): void {
+  private collapseUnusedBlocks(usedNodes: Iterable<TurboshaftGraphOperation>): void {
     const usedBlocks = new Set<TurboshaftGraphBlock>();
     for (const node of usedNodes) {
       usedBlocks.add(node.block);
@@ -885,7 +910,7 @@ export class TurboshaftGraphView extends MovableView<TurboshaftGraph> {
   }
 
   private selectNodesOfSelectedBlocks(): void {
-    let selectedNodes = new Array<TurboshaftGraphNode>();
+    let selectedNodes = new Array<TurboshaftGraphOperation>();
     for (const key of this.state.blocksSelection.selectedKeys()) {
       const block = this.graph.blockMap[key];
       if (!block) continue;
