@@ -457,7 +457,7 @@ function clearDefaultTriggerAsyncId() {
  * @template {unknown} R
  * @param {number} triggerAsyncId
  * @param { (...T: args) => R } block
- * @param  {T} args
+ * @param {T} args
  * @returns {R}
  */
 function defaultTriggerAsyncIdScope(triggerAsyncId, block, ...args) {
@@ -532,7 +532,7 @@ function emitAfterScript(asyncId) {
 
 function emitDestroyScript(asyncId) {
   // Return early if there are no destroy callbacks, or invalid asyncId.
-  if (!hasHooks(kDestroy) || asyncId <= 0)
+  if (!hasHooks(kDestroy) || !(asyncId > 0))
     return;
   async_wrap.queueDestroyAsyncId(asyncId);
 }
