@@ -5,8 +5,8 @@ const {
 } = primordials;
 
 const {
-  prepareMainThreadExecution,
   markBootstrapComplete,
+  prepareTestRunnerMainExecution,
 } = require('internal/process/pre_execution');
 const { isUsingInspector } = require('internal/util/inspector');
 const { run } = require('internal/test_runner/runner');
@@ -16,10 +16,12 @@ let debug = require('internal/util/debuglog').debuglog('test_runner', (fn) => {
   debug = fn;
 });
 
-prepareMainThreadExecution(false);
-markBootstrapComplete();
-
 const options = parseCommandLine();
+const isTestIsolationDisabled = options.isolation === 'none';
+// We set initializeModules to false as we want to load user modules in the test runner run function
+// if we are running with --test-isolation=none
+prepareTestRunnerMainExecution(!isTestIsolationDisabled);
+markBootstrapComplete();
 
 if (isUsingInspector() && options.isolation === 'process') {
   process.emitWarning('Using the inspector with --test forces running at a concurrency of 1. ' +
