@@ -44,6 +44,7 @@ const {
   findLongestRegisteredExtension,
   resolveForCJSWithHooks,
   loadSourceForCJSWithHooks,
+  populateCJSExportsFromESM,
 } = require('internal/modules/cjs/loader');
 const { fileURLToPath, pathToFileURL, URL } = require('internal/url');
 let debug = require('internal/util/debuglog').debuglog('esm', (fn) => {
@@ -88,6 +89,7 @@ exports.translators = translators;
 /**
  * Converts a URL to a file path if the URL protocol is 'file:'.
  * @param {string} url - The URL to convert.
+ * @returns {string|URL}
  */
 function errPath(url) {
   const parsed = new URL(url);
@@ -98,10 +100,12 @@ function errPath(url) {
 }
 
 // Strategy for loading a standard JavaScript module.
-translators.set('module', function moduleStrategy(url, source, isMain) {
+translators.set('module', function moduleStrategy(url, translateContext, parentURL) {
+  let { source } = translateContext;
+  const isMain = (parentURL === undefined);
   assertBufferSource(source, true, 'load');
   source = stringify(source);
-  debug(`Translating StandardModule ${url}`);
+  debug(`Translating StandardModule ${url}`, translateContext);
   const { compileSourceTextModule } = require('internal/modules/esm/utils');
   const context = isMain ? { isMain } : undefined;
   const module = compileSourceTextModule(url, source, this, context);
@@ -147,9 +151,32 @@ function loadCJSModule(module, source, url, filename, isMain) {
       }
       specifier = `${pathToFileURL(path)}`;
     }
+
+    // FIXME(node:59666) Currently, the ESM loader re-invents require() here for imported CJS and this
+    // requires a separate cache to be populated as well as introducing several quirks. This is not ideal.
     const job = cascadedLoader.getModuleJobForRequireInImportedCJS(specifier, url, importAttributes);
     job.runSync();
-    return cjsCache.get(job.url).exports;
+    let mod = cjsCache.get(job.url);
+    assert(job.module, `Imported CJS module ${url} failed to load module ${job.url} using require() due to race condition`);
+
+    if (job.module.synthetic) {
+      assert(mod, `Imported CJS module ${url} failed to load module ${job.url} using require() due to missed cache`);
+      return mod.exports;
+    }
+
+    // The module being required is a source text module.
+    if (!mod) {
+      mod = cjsEmplaceModuleCacheEntry(job.url);
+      cjsCache.set(job.url, mod);
+    }
+    // The module has been cached by the re-invented require. Update the exports object
+    // from the namespace object and return the evaluated exports.
+    if (!mod.loaded) {
+      debug('populateCJSExportsFromESM from require(esm) in imported CJS', url, mod, job.module);
+      populateCJSExportsFromESM(mod, job.module, job.module.getNamespace());
+      mod.loaded = true;
+    }
+    return mod.exports;
   };
   setOwnProperty(requireFn, 'resolve', function resolve(specifier) {
     if (!StringPrototypeStartsWith(specifier, 'node:')) {
@@ -170,32 +197,42 @@ function loadCJSModule(module, source, url, filename, isMain) {
 
 // TODO: can we use a weak map instead?
 const cjsCache = new SafeMap();
+
 /**
  * Creates a ModuleWrap object for a CommonJS module.
  * @param {string} url - The URL of the module.
- * @param {string} source - The source code of the module.
- * @param {boolean} isMain - Whether the module is the main module.
- * @param {string} format - Format of the module.
- * @param {typeof loadCJSModule} [loadCJS=loadCJSModule] - The function to load the CommonJS module.
+ * @param {{ format: ModuleFormat, source: ModuleSource }} translateContext Context for the translator
+ * @param {string|undefined} parentURL URL of the module initiating the module loading for the first time.
+ *   Undefined if it's the entry point.
+ * @param {typeof loadCJSModule} [loadCJS] - The function to load the CommonJS module.
  * @returns {ModuleWrap} The ModuleWrap object for the CommonJS module.
  */
-function createCJSModuleWrap(url, source, isMain, format, loadCJS = loadCJSModule) {
-  debug(`Translating CJSModule ${url}`);
+function createCJSModuleWrap(url, translateContext, parentURL, loadCJS = loadCJSModule) {
+  debug(`Translating CJSModule ${url}`, translateContext);
 
+  const { format: sourceFormat } = translateContext;
+  let { source } = translateContext;
+  const isMain = (parentURL === undefined);
   const filename = urlToFilename(url);
   // In case the source was not provided by the `load` step, we need fetch it now.
   source = stringify(source ?? getSource(new URL(url)).source);
 
-  const { exportNames, module } = cjsPreparseModuleExports(filename, source, format);
+  const { exportNames, module } = cjsPreparseModuleExports(filename, source, sourceFormat);
   cjsCache.set(url, module);
-  const namesWithDefault = exportNames.has('default') ?
-    [...exportNames] : ['default', ...exportNames];
+
+  const wrapperNames = [...exportNames];
+  if (!exportNames.has('default')) {
+    ArrayPrototypePush(wrapperNames, 'default');
+  }
+  if (!exportNames.has('module.exports')) {
+    ArrayPrototypePush(wrapperNames, 'module.exports');
+  }
 
   if (isMain) {
     setOwnProperty(process, 'mainModule', module);
   }
 
-  return new ModuleWrap(url, undefined, namesWithDefault, function() {
+  return new ModuleWrap(url, undefined, wrapperNames, function() {
     debug(`Loading CJSModule ${url}`);
 
     if (!module.loaded) {
@@ -210,8 +247,8 @@ function createCJSModuleWrap(url, source, isMain, format, loadCJS = loadCJSModul
       ({ exports } = module);
     }
     for (const exportName of exportNames) {
-      if (!ObjectPrototypeHasOwnProperty(exports, exportName) ||
-          exportName === 'default') {
+      if (exportName === 'default' || exportName === 'module.exports' ||
+          !ObjectPrototypeHasOwnProperty(exports, exportName)) {
         continue;
       }
       // We might trigger a getter -> dont fail.
@@ -224,17 +261,19 @@ function createCJSModuleWrap(url, source, isMain, format, loadCJS = loadCJSModul
       this.setExport(exportName, value);
     }
     this.setExport('default', exports);
+    this.setExport('module.exports', exports);
   }, module);
 }
 
 /**
  * Creates a ModuleWrap object for a CommonJS module without source texts.
  * @param {string} url - The URL of the module.
- * @param {boolean} isMain - Whether the module is the main module.
+ * @param {string|undefined} parentURL - URL of the parent module, if any.
  * @returns {ModuleWrap} The ModuleWrap object for the CommonJS module.
  */
-function createCJSNoSourceModuleWrap(url, isMain) {
+function createCJSNoSourceModuleWrap(url, parentURL) {
   debug(`Translating CJSModule without source ${url}`);
+  const isMain = (parentURL === undefined);
 
   const filename = urlToFilename(url);
 
@@ -268,69 +307,76 @@ function createCJSNoSourceModuleWrap(url, isMain) {
   }, module);
 }
 
-translators.set('commonjs-sync', function requireCommonJS(url, source, isMain) {
+translators.set('commonjs-sync', function requireCommonJS(url, translateContext, parentURL) {
   initCJSParseSync();
 
-  return createCJSModuleWrap(url, source, isMain, 'commonjs', (module, source, url, filename, isMain) => {
-    assert(module === CJSModule._cache[filename]);
-    wrapModuleLoad(filename, null, isMain);
-  });
+  return createCJSModuleWrap(url, translateContext, parentURL, loadCJSModuleWithModuleLoad);
 });
 
 // Handle CommonJS modules referenced by `require` calls.
 // This translator function must be sync, as `require` is sync.
-translators.set('require-commonjs', (url, source, isMain) => {
+translators.set('require-commonjs', (url, translateContext, parentURL) => {
   initCJSParseSync();
   assert(cjsParse);
 
-  return createCJSModuleWrap(url, source, isMain, 'commonjs');
+  return createCJSModuleWrap(url, translateContext, parentURL);
 });
 
 // Handle CommonJS modules referenced by `require` calls.
 // This translator function must be sync, as `require` is sync.
-translators.set('require-commonjs-typescript', (url, source, isMain) => {
+translators.set('require-commonjs-typescript', (url, translateContext, parentURL) => {
   assert(cjsParse);
-  const code = stripTypeScriptModuleTypes(stringify(source), url);
-  return createCJSModuleWrap(url, code, isMain, 'commonjs-typescript');
+  translateContext.source = stripTypeScriptModuleTypes(stringify(translateContext.source), url);
+  return createCJSModuleWrap(url, translateContext, parentURL);
 });
 
+// This goes through Module._load to accommodate monkey-patchers.
+function loadCJSModuleWithModuleLoad(module, source, url, filename, isMain) {
+  assert(module === CJSModule._cache[filename]);
+  wrapModuleLoad(filename, undefined, isMain);
+}
+
 // Handle CommonJS modules referenced by `import` statements or expressions,
 // or as the initial entry point when the ESM loader handles a CommonJS entry.
-translators.set('commonjs', function commonjsStrategy(url, source, isMain) {
+translators.set('commonjs', function commonjsStrategy(url, translateContext, parentURL) {
   if (!cjsParse) {
     initCJSParseSync();
   }
 
   // For backward-compatibility, it's possible to return a nullish value for
-  // CJS source associated with a file: URL. In this case, the source is
-  // obtained by calling the monkey-patchable CJS loader.
-  const cjsLoader = source == null ? (module, source, url, filename, isMain) => {
-    assert(module === CJSModule._cache[filename]);
-    wrapModuleLoad(filename, undefined, isMain);
-  } : loadCJSModule;
+  // CJS source associated with a `file:` URL - that usually means the source is not
+  // customized (is loaded by default load) or the hook author wants it to be reloaded
+  // through CJS routine. In this case, the source is obtained by calling the
+  // monkey-patchable CJS loader.
+  // TODO(joyeecheung): just use wrapModuleLoad and let the CJS loader
+  // invoke the off-thread hooks. Use a special parent to avoid invoking in-thread
+  // hooks twice.
+  const shouldReloadByCJSLoader = (translateContext.shouldBeReloadedByCJSLoader || translateContext.source == null);
+  const cjsLoader = shouldReloadByCJSLoader ? loadCJSModuleWithModuleLoad : loadCJSModule;
 
   try {
     // We still need to read the FS to detect the exports.
-    source ??= readFileSync(new URL(url), 'utf8');
+    translateContext.source ??= readFileSync(new URL(url), 'utf8');
   } catch {
     // Continue regardless of error.
   }
-  return createCJSModuleWrap(url, source, isMain, 'commonjs', cjsLoader);
+  return createCJSModuleWrap(url, translateContext, parentURL, cjsLoader);
 });
 
 /**
  * Get or create an entry in the CJS module cache for the given filename.
  * @param {string} filename CJS module filename
+ * @param {CJSModule} parent The parent CJS module
  * @returns {CJSModule} the cached CJS module entry
  */
-function cjsEmplaceModuleCacheEntry(filename, exportNames) {
+function cjsEmplaceModuleCacheEntry(filename, parent) {
   // TODO: Do we want to keep hitting the user mutable CJS loader here?
   let cjsMod = CJSModule._cache[filename];
   if (cjsMod) {
     return cjsMod;
   }
 
-  cjsMod = new CJSModule(filename);
+  cjsMod = new CJSModule(filename, parent);
   cjsMod.filename = filename;
   cjsMod.paths = CJSModule._nodeModulePaths(cjsMod.path);
   cjsMod[kIsCachedByESMLoader] = true;
@@ -344,6 +390,7 @@ function cjsEmplaceModuleCacheEntry(filename, exportNames) {
  * @param {string} filename - The filename of the module.
  * @param {string} [source] - The source code of the module.
  * @param {string} [format]
+ * @returns {{module: CJSModule, exportNames: string[]}}
  */
 function cjsPreparseModuleExports(filename, source, format) {
   const module = cjsEmplaceModuleCacheEntry(filename);
@@ -403,8 +450,8 @@ function cjsPreparseModuleExports(filename, source, format) {
 
 // Strategy for loading a node builtin CommonJS module that isn't
 // through normal resolution
-translators.set('builtin', function builtinStrategy(url) {
-  debug(`Translating BuiltinModule ${url}`);
+translators.set('builtin', function builtinStrategy(url, translateContext) {
+  debug(`Translating BuiltinModule ${url}`, translateContext);
   // Slice 'node:' scheme
   const id = StringPrototypeSlice(url, 5);
   const module = loadBuiltinModule(id, url);
@@ -417,7 +464,8 @@ translators.set('builtin', function builtinStrategy(url) {
 });
 
 // Strategy for loading a JSON file
-translators.set('json', function jsonStrategy(url, source) {
+translators.set('json', function jsonStrategy(url, translateContext) {
+  let { source } = translateContext;
   assertBufferSource(source, true, 'load');
   debug(`Loading JSONModule ${url}`);
   const pathname = StringPrototypeStartsWith(url, 'file:') ?
@@ -485,10 +533,11 @@ translators.set('json', function jsonStrategy(url, source) {
  * >} [[Instance]] slot proxy for WebAssembly Module Record
  */
 const wasmInstances = new SafeWeakMap();
-translators.set('wasm', function(url, source) {
+translators.set('wasm', function(url, translateContext) {
+  const { source } = translateContext;
   assertBufferSource(source, false, 'load');
 
-  debug(`Translating WASMModule ${url}`);
+  debug(`Translating WASMModule ${url}`, translateContext);
 
   let compiled;
   try {
@@ -548,7 +597,6 @@ translators.set('wasm', function(url, source) {
         reflect.imports[impt] = wrappedModule;
       }
     }
-
     // In cycles importing unexecuted Wasm, wasmInstance will be undefined, which will fail during
     // instantiation, since all bindings will be in the Temporal Deadzone (TDZ).
     const { exports } = new WebAssembly.Instance(compiled, reflect.imports);
@@ -569,13 +617,17 @@ translators.set('wasm', function(url, source) {
       reflect.exports[expt].set(val);
     }
   });
+  // WebAssembly modules support source phase imports, to import the compiled module
+  // separate from the linked instance.
+  module.setModuleSourceObject(compiled);
   return module;
 });
 
 // Strategy for loading a addon
-translators.set('addon', function translateAddon(url, source, isMain) {
+translators.set('addon', function translateAddon(url, translateContext, parentURL) {
   emitExperimentalWarning('Importing addons');
 
+  const { source } = translateContext;
   // The addon must be loaded from file system with dlopen. Assert
   // the source is null.
   if (source !== null) {
@@ -586,23 +638,25 @@ translators.set('addon', function translateAddon(url, source, isMain) {
       source);
   }
 
-  debug(`Translating addon ${url}`);
+  debug(`Translating addon ${url}`, translateContext);
 
-  return createCJSNoSourceModuleWrap(url, isMain);
+  return createCJSNoSourceModuleWrap(url, parentURL);
 });
 
 // Strategy for loading a commonjs TypeScript module
-translators.set('commonjs-typescript', function(url, source, isMain) {
+translators.set('commonjs-typescript', function(url, translateContext, parentURL) {
+  const { source } = translateContext;
   assertBufferSource(source, true, 'load');
-  const code = stripTypeScriptModuleTypes(stringify(source), url);
-  debug(`Translating TypeScript ${url}`);
-  return FunctionPrototypeCall(translators.get('commonjs'), this, url, code, isMain);
+  debug(`Translating TypeScript ${url}`, translateContext);
+  translateContext.source = stripTypeScriptModuleTypes(stringify(source), url);
+  return FunctionPrototypeCall(translators.get('commonjs'), this, url, translateContext, parentURL);
 });
 
 // Strategy for loading an esm TypeScript module
-translators.set('module-typescript', function(url, source, isMain) {
+translators.set('module-typescript', function(url, translateContext, parentURL) {
+  const { source } = translateContext;
   assertBufferSource(source, true, 'load');
-  const code = stripTypeScriptModuleTypes(stringify(source), url);
-  debug(`Translating TypeScript ${url}`);
-  return FunctionPrototypeCall(translators.get('module'), this, url, code, isMain);
+  debug(`Translating TypeScript ${url}`, translateContext);
+  translateContext.source = stripTypeScriptModuleTypes(stringify(source), url);
+  return FunctionPrototypeCall(translators.get('module'), this, url, translateContext, parentURL);
 });
