@@ -16,7 +16,6 @@ const {
   ObjectSetPrototypeOf,
   ReflectGetOwnPropertyDescriptor,
   ReflectOwnKeys,
-  RegExpPrototypeSymbolReplace,
   SafeMap,
   SafeSet,
   StringPrototypeCharAt,
@@ -30,9 +29,13 @@ const {
   Symbol,
   SymbolIterator,
   SymbolToStringTag,
+  TypedArrayPrototypeGetBuffer,
+  TypedArrayPrototypeGetByteLength,
+  TypedArrayPrototypeGetByteOffset,
   decodeURIComponent,
 } = primordials;
 
+const { URLPattern } = internalBinding('url_pattern');
 const { inspect } = require('internal/util/inspect');
 const {
   encodeStr,
@@ -46,8 +49,17 @@ const {
   kEnumerableProperty,
   kEmptyObject,
   SideEffectFreeRegExpPrototypeSymbolReplace,
+  isWindows,
 } = require('internal/util');
 
+const {
+  platform,
+} = require('internal/process/per_thread');
+
+const {
+  markTransferMode,
+} = require('internal/worker/js_transferable');
+
 const {
   codes: {
     ERR_ARG_NOT_ITERABLE,
@@ -72,17 +84,20 @@ const {
   CHAR_LOWERCASE_Z,
   CHAR_PERCENT,
   CHAR_PLUS,
+  CHAR_COLON,
 } = require('internal/constants');
 const path = require('path');
+const { Buffer } = require('buffer');
 
 const {
   validateFunction,
+  validateObject,
+  kValidateObjectAllowObjects,
 } = require('internal/validators');
 
-const querystring = require('querystring');
+const { percentDecode } = require('internal/data_url');
 
-const { platform } = process;
-const isWindows = platform === 'win32';
+const querystring = require('querystring');
 
 const bindingUrl = internalBinding('url');
 
@@ -328,6 +343,8 @@ class URLSearchParams {
   // Default parameter is necessary to keep URLSearchParams.length === 0 in
   // accordance with Web IDL spec.
   constructor(init = undefined) {
+    markTransferMode(this, false, false);
+
     if (init == null) {
       // Do nothing
     } else if (typeof init === 'object' || typeof init === 'function') {
@@ -771,6 +788,8 @@ function isURL(self) {
  * for invalid URL inputs.
  */
 const kParseURLSymbol = Symbol('kParseURL');
+const kCreateURLFromPosixPathSymbol = Symbol('kCreateURLFromPosixPath');
+const kCreateURLFromWindowsPathSymbol = Symbol('kCreateURLFromWindowsPath');
 
 class URL {
   #context = new URLContext();
@@ -791,6 +810,8 @@ class URL {
   }
 
   constructor(input, base = undefined, parseSymbol = undefined) {
+    markTransferMode(this, false, false);
+
     if (arguments.length === 0) {
       throw new ERR_MISSING_ARGS('url');
     }
@@ -802,8 +823,17 @@ class URL {
       base = `${base}`;
     }
 
-    const raiseException = parseSymbol !== kParseURLSymbol;
-    const href = bindingUrl.parse(input, base, raiseException);
+    let href;
+    if (arguments.length < 3) {
+      href = bindingUrl.parse(input, base, true);
+    } else {
+      const raiseException = parseSymbol !== kParseURLSymbol;
+      const interpretAsWindowsPath = parseSymbol === kCreateURLFromWindowsPathSymbol;
+      const pathToFileURL = interpretAsWindowsPath || (parseSymbol === kCreateURLFromPosixPathSymbol);
+      href = pathToFileURL ?
+        bindingUrl.pathToFileURL(input, interpretAsWindowsPath, base) :
+        bindingUrl.parse(input, base, raiseException);
+    }
     if (href) {
       this.#updateContext(href);
     }
@@ -1403,6 +1433,7 @@ function domainToUnicode(domain) {
  * @returns {Record<string, unknown>}
  */
 function urlToHttpOptions(url) {
+  validateObject(url, 'url', kValidateObjectAllowObjects);
   const { hostname, pathname, port, username, password, search } = url;
   const options = {
     __proto__: null,
@@ -1435,7 +1466,7 @@ function getPathFromURLWin32(url) {
       if ((pathname[n + 1] === '2' && third === 102) || // 2f 2F /
           (pathname[n + 1] === '5' && third === 99)) {  // 5c 5C \
         throw new ERR_INVALID_FILE_URL_PATH(
-          'must not include encoded \\ or / characters',
+          'must not include encoded \\ or / characters', url,
         );
       }
     }
@@ -1456,11 +1487,81 @@ function getPathFromURLWin32(url) {
   const sep = StringPrototypeCharAt(pathname, 2);
   if (letter < CHAR_LOWERCASE_A || letter > CHAR_LOWERCASE_Z ||   // a..z A..Z
       (sep !== ':')) {
-    throw new ERR_INVALID_FILE_URL_PATH('must be absolute');
+    throw new ERR_INVALID_FILE_URL_PATH('must be absolute', url);
   }
   return StringPrototypeSlice(pathname, 1);
 }
 
+function getPathBufferFromURLWin32(url) {
+  const hostname = url.hostname;
+  let pathname = url.pathname;
+  // In the getPathFromURLWin32 variant, we scan the input for backslash (\)
+  // and forward slash (/) characters, specifically looking for the ASCII/UTF8
+  // encoding these and forbidding their use. This is a bit tricky
+  // because these may conflict with non-UTF8 encodings. For instance,
+  // in shift-jis, %5C identifies the symbol for the Japanese Yen and not the
+  // backslash. If we have a url like file:///foo/%5c/bar, then we really have
+  // no way of knowing if that %5c is meant to be a backslash \ or a yen sign.
+  // Passing in an encoding option does not help since our Buffer encoding only
+  // knows about certain specific text encodings and a single file path might
+  // actually contain segments that use multiple encodings. It's tricky! So,
+  // for this variation where we are producing a buffer, we won't scan for the
+  // slashes at all, and instead will decode the bytes literally into the
+  // returned Buffer. That said, that can also be tricky because, on windows,
+  // the file path separator *is* the ASCII backslash. This is a known issue
+  // on windows specific to the Shift-JIS encoding that we're not really going
+  // to solve here. Instead, we're going to do the best we can and just
+  // interpret the input url as a sequence of bytes.
+
+  // Because we are converting to a Windows file path here, we need to replace
+  // the explicit forward slash separators with backslashes. Note that this
+  // intentionally disregards any percent-encoded forward slashes in the path.
+  pathname = SideEffectFreeRegExpPrototypeSymbolReplace(FORWARD_SLASH, pathname, '\\');
+
+  // Now, let's start to build our Buffer. We will initially start with a
+  // Buffer allocated to fit in the entire string. Worst case there are no
+  // percent encoded characters and we take the string as is. Any invalid
+  // percent encodings, e.g. `%ZZ` are ignored and are passed through
+  // literally.
+  const decodedu8 = percentDecode(Buffer.from(pathname, 'utf8'));
+  const decodedPathname = Buffer.from(TypedArrayPrototypeGetBuffer(decodedu8),
+                                      TypedArrayPrototypeGetByteOffset(decodedu8),
+                                      TypedArrayPrototypeGetByteLength(decodedu8));
+  if (hostname !== '') {
+    // If hostname is set, then we have a UNC path
+    // Pass the hostname through domainToUnicode just in case
+    // it is an IDN using punycode encoding. We do not need to worry
+    // about percent encoding because the URL parser will have
+    // already taken care of that for us. Note that this only
+    // causes IDNs with an appropriate `xn--` prefix to be decoded.
+
+    // This is a bit tricky because of the need to convert to a Buffer
+    // followed by concatenation of the results.
+    const prefix = Buffer.from('\\\\', 'ascii');
+    const domain = Buffer.from(domainToUnicode(hostname), 'utf8');
+
+    return Buffer.concat([prefix, domain, decodedPathname]);
+  }
+  // Otherwise, it's a local path that requires a drive letter
+  // In this case we're only going to pay attention to the second and
+  // third bytes in the decodedPathname. If first byte is either an ASCII
+  // uppercase letter between 'A' and 'Z' or lowercase letter between
+  // 'a' and 'z', and the second byte must be an ASCII `:` or the
+  // operation will fail.
+
+  const letter = decodedPathname[1] | 0x20;
+  const sep = decodedPathname[2];
+
+  if (letter < CHAR_LOWERCASE_A || letter > CHAR_LOWERCASE_Z ||   // a..z A..Z
+      (sep !== CHAR_COLON)) {
+    throw new ERR_INVALID_FILE_URL_PATH('must be absolute', url);
+  }
+
+  // Now, we'll just return everything except the first byte of
+  // decodedPathname
+  return decodedPathname.subarray(1);
+}
+
 function getPathFromURLPosix(url) {
   if (url.hostname !== '') {
     throw new ERR_INVALID_FILE_URL_HOST(platform);
@@ -1472,6 +1573,7 @@ function getPathFromURLPosix(url) {
       if (pathname[n + 1] === '2' && third === 102) {
         throw new ERR_INVALID_FILE_URL_PATH(
           'must not include encoded / characters',
+          url,
         );
       }
     }
@@ -1479,6 +1581,28 @@ function getPathFromURLPosix(url) {
   return decodeURIComponent(pathname);
 }
 
+function getPathBufferFromURLPosix(url) {
+  if (url.hostname !== '') {
+    throw new ERR_INVALID_FILE_URL_HOST(platform);
+  }
+  const pathname = url.pathname;
+
+  // In the getPathFromURLPosix variant, we scan the input for forward slash
+  // (/) characters, specifically looking for the ASCII/UTF8 and forbidding
+  // its use. This is a bit tricky because these may conflict with non-UTF8
+  // encodings. Passing in an encoding option does not help since our Buffer
+  // encoding only knows about certain specific text encodings and a single
+  // file path might actually contain segments that use multiple encodings.
+  // It's tricky! So, for this variation where we are producing a buffer, we
+  // won't scan for the slashes at all, and instead will decode the bytes
+  // literally into the returned Buffer. We're going to do the best we can and
+  // just interpret the input url as a sequence of bytes.
+  const u8 = percentDecode(Buffer.from(pathname, 'utf8'));
+  return Buffer.from(TypedArrayPrototypeGetBuffer(u8),
+                     TypedArrayPrototypeGetByteOffset(u8),
+                     TypedArrayPrototypeGetByteLength(u8));
+}
+
 function fileURLToPath(path, options = kEmptyObject) {
   const windows = options?.windows;
   if (typeof path === 'string')
@@ -1490,106 +1614,55 @@ function fileURLToPath(path, options = kEmptyObject) {
   return (windows ?? isWindows) ? getPathFromURLWin32(path) : getPathFromURLPosix(path);
 }
 
-// RFC1738 defines the following chars as "unsafe" for URLs
-// @see https://www.ietf.org/rfc/rfc1738.txt 2.2. URL Character Encoding Issues
-const percentRegEx = /%/g;
-const newlineRegEx = /\n/g;
-const carriageReturnRegEx = /\r/g;
-const tabRegEx = /\t/g;
-const quoteRegEx = /"/g;
-const hashRegex = /#/g;
-const spaceRegEx = / /g;
-const questionMarkRegex = /\?/g;
-const openSquareBracketRegEx = /\[/g;
-const backslashRegEx = /\\/g;
-const closeSquareBracketRegEx = /]/g;
-const caretRegEx = /\^/g;
-const verticalBarRegEx = /\|/g;
-const tildeRegEx = /~/g;
-
-function encodePathChars(filepath, options = kEmptyObject) {
-  if (StringPrototypeIncludes(filepath, '%')) {
-    filepath = RegExpPrototypeSymbolReplace(percentRegEx, filepath, '%25');
+// An alternative to fileURLToPath that outputs a Buffer
+// instead of a string. The other fileURLToPath does not
+// handle non-UTF8 encoded percent encodings at all, so
+// converting to a Buffer is necessary in cases where the
+// to string conversion would fail.
+function fileURLToPathBuffer(path, options = kEmptyObject) {
+  const windows = options?.windows;
+  if (typeof path === 'string') {
+    path = new URL(path);
+  } else if (!isURL(path)) {
+    throw new ERR_INVALID_ARG_TYPE('path', ['string', 'URL'], path);
   }
-
-  if (StringPrototypeIncludes(filepath, '\t')) {
-    filepath = RegExpPrototypeSymbolReplace(tabRegEx, filepath, '%09');
+  if (path.protocol !== 'file:') {
+    throw new ERR_INVALID_URL_SCHEME('file');
   }
-  if (StringPrototypeIncludes(filepath, '\n')) {
-    filepath = RegExpPrototypeSymbolReplace(newlineRegEx, filepath, '%0A');
-  }
-  if (StringPrototypeIncludes(filepath, '\r')) {
-    filepath = RegExpPrototypeSymbolReplace(carriageReturnRegEx, filepath, '%0D');
-  }
-  if (StringPrototypeIncludes(filepath, ' ')) {
-    filepath = RegExpPrototypeSymbolReplace(spaceRegEx, filepath, '%20');
-  }
-  if (StringPrototypeIncludes(filepath, '"')) {
-    filepath = RegExpPrototypeSymbolReplace(quoteRegEx, filepath, '%22');
-  }
-  if (StringPrototypeIncludes(filepath, '#')) {
-    filepath = RegExpPrototypeSymbolReplace(hashRegex, filepath, '%23');
-  }
-  if (StringPrototypeIncludes(filepath, '?')) {
-    filepath = RegExpPrototypeSymbolReplace(questionMarkRegex, filepath, '%3F');
-  }
-  if (StringPrototypeIncludes(filepath, '[')) {
-    filepath = RegExpPrototypeSymbolReplace(openSquareBracketRegEx, filepath, '%5B');
-  }
-  // Back-slashes must be special-cased on Windows, where they are treated as path separator.
-  if (!options.windows && StringPrototypeIncludes(filepath, '\\')) {
-    filepath = RegExpPrototypeSymbolReplace(backslashRegEx, filepath, '%5C');
-  }
-  if (StringPrototypeIncludes(filepath, ']')) {
-    filepath = RegExpPrototypeSymbolReplace(closeSquareBracketRegEx, filepath, '%5D');
-  }
-  if (StringPrototypeIncludes(filepath, '^')) {
-    filepath = RegExpPrototypeSymbolReplace(caretRegEx, filepath, '%5E');
-  }
-  if (StringPrototypeIncludes(filepath, '|')) {
-    filepath = RegExpPrototypeSymbolReplace(verticalBarRegEx, filepath, '%7C');
-  }
-  if (StringPrototypeIncludes(filepath, '~')) {
-    filepath = RegExpPrototypeSymbolReplace(tildeRegEx, filepath, '%7E');
-  }
-
-  return filepath;
+  return (windows ?? isWindows) ? getPathBufferFromURLWin32(path) : getPathBufferFromURLPosix(path);
 }
 
 function pathToFileURL(filepath, options = kEmptyObject) {
   const windows = options?.windows ?? isWindows;
-  if (windows && StringPrototypeStartsWith(filepath, '\\\\')) {
-    const outURL = new URL('file://');
+  const isUNC = windows && StringPrototypeStartsWith(filepath, '\\\\');
+  let resolved = isUNC ?
+    filepath :
+    (windows ? path.win32.resolve(filepath) : path.posix.resolve(filepath));
+  if (isUNC || (windows && StringPrototypeStartsWith(resolved, '\\\\'))) {
     // UNC path format: \\server\share\resource
     // Handle extended UNC path and standard UNC path
     // "\\?\UNC\" path prefix should be ignored.
     // Ref: https://learn.microsoft.com/en-us/windows/win32/fileio/maximum-file-path-limitation
-    const isExtendedUNC = StringPrototypeStartsWith(filepath, '\\\\?\\UNC\\');
+    const isExtendedUNC = StringPrototypeStartsWith(resolved, '\\\\?\\UNC\\');
     const prefixLength = isExtendedUNC ? 8 : 2;
-    const hostnameEndIndex = StringPrototypeIndexOf(filepath, '\\', prefixLength);
+    const hostnameEndIndex = StringPrototypeIndexOf(resolved, '\\', prefixLength);
     if (hostnameEndIndex === -1) {
       throw new ERR_INVALID_ARG_VALUE(
         'path',
-        filepath,
+        resolved,
         'Missing UNC resource path',
       );
     }
     if (hostnameEndIndex === 2) {
       throw new ERR_INVALID_ARG_VALUE(
         'path',
-        filepath,
+        resolved,
         'Empty UNC servername',
       );
     }
-    const hostname = StringPrototypeSlice(filepath, prefixLength, hostnameEndIndex);
-    outURL.hostname = domainToASCII(hostname);
-    outURL.pathname = encodePathChars(
-      RegExpPrototypeSymbolReplace(backslashRegEx, StringPrototypeSlice(filepath, hostnameEndIndex), '/'),
-      { windows },
-    );
-    return outURL;
+    const hostname = StringPrototypeSlice(resolved, prefixLength, hostnameEndIndex);
+    return new URL(StringPrototypeSlice(resolved, hostnameEndIndex), hostname, kCreateURLFromWindowsPathSymbol);
   }
-  let resolved = windows ? path.win32.resolve(filepath) : path.posix.resolve(filepath);
   // path.resolve strips trailing slashes so we must add them back
   const filePathLast = StringPrototypeCharCodeAt(filepath,
                                                  filepath.length - 1);
@@ -1598,7 +1671,7 @@ function pathToFileURL(filepath, options = kEmptyObject) {
       resolved[resolved.length - 1] !== path.sep)
     resolved += '/';
 
-  return new URL(`file://${encodePathChars(resolved, { windows })}`);
+  return new URL(resolved, undefined, windows ? kCreateURLFromWindowsPathSymbol : kCreateURLFromPosixPathSymbol);
 }
 
 function toPathIfFileURL(fileURLOrPath) {
@@ -1619,10 +1692,12 @@ function getURLOrigin(url) {
 
 module.exports = {
   fileURLToPath,
+  fileURLToPathBuffer,
   pathToFileURL,
   toPathIfFileURL,
   installObjectURLMethods,
   URL,
+  URLPattern,
   URLSearchParams,
   URLParse: URL.parse,
   domainToASCII,
