@@ -19,6 +19,7 @@ namespace node {
 using ncrypto::BignumPointer;
 using ncrypto::Cipher;
 using ncrypto::CipherCtxPointer;
+using ncrypto::DataPointer;
 using v8::FunctionCallbackInfo;
 using v8::Just;
 using v8::JustVoid;
@@ -31,6 +32,9 @@ using v8::Value;
 
 namespace crypto {
 namespace {
+constexpr size_t kAesBlockSize = 16;
+constexpr const char* kDefaultWrapIV = "\xa6\xa6\xa6\xa6\xa6\xa6\xa6\xa6";
+
 // Implements general AES encryption and decryption for CBC
 // The key_data must be a secret key.
 // On success, this function sets out to a new ByteSource
@@ -43,11 +47,11 @@ WebCryptoCipherStatus AES_Cipher(Environment* env,
                                  ByteSource* out) {
   CHECK_EQ(key_data.GetKeyType(), kKeyTypeSecret);
 
-  const int mode = params.cipher.getMode();
-
   auto ctx = CipherCtxPointer::New();
-  if (mode == EVP_CIPH_WRAP_MODE) {
-    ctx.setFlags(EVP_CIPHER_CTX_FLAG_WRAP_ALLOW);
+  CHECK(ctx);
+
+  if (params.cipher.isWrapMode()) {
+    ctx.setAllowWrap();
   }
 
   const bool encrypt = cipher_mode == kWebCryptoCipherEncrypt;
@@ -57,7 +61,8 @@ WebCryptoCipherStatus AES_Cipher(Environment* env,
     return WebCryptoCipherStatus::FAILED;
   }
 
-  if (mode == EVP_CIPH_GCM_MODE && !ctx.setIvLength(params.iv.size())) {
+  if ((params.cipher.isGcmMode() || params.cipher.isOcbMode()) &&
+      !ctx.setIvLength(params.iv.size())) {
     return WebCryptoCipherStatus::FAILED;
   }
 
@@ -72,11 +77,20 @@ WebCryptoCipherStatus AES_Cipher(Environment* env,
 
   size_t tag_len = 0;
 
-  if (mode == EVP_CIPH_GCM_MODE) {
+  if (params.cipher.isGcmMode() || params.cipher.isOcbMode()) {
     switch (cipher_mode) {
       case kWebCryptoCipherDecrypt: {
         // If in decrypt mode, the auth tag must be set in the params.tag.
         CHECK(params.tag);
+
+        // For OCB mode, we need to set the auth tag length before setting the
+        // tag
+        if (params.cipher.isOcbMode()) {
+          if (!ctx.setAeadTagLength(params.tag.size())) {
+            return WebCryptoCipherStatus::FAILED;
+          }
+        }
+
         ncrypto::Buffer<const char> buffer = {
             .data = params.tag.data<char>(),
             .len = params.tag.size(),
@@ -87,12 +101,19 @@ WebCryptoCipherStatus AES_Cipher(Environment* env,
         break;
       }
       case kWebCryptoCipherEncrypt: {
-        // In decrypt mode, we grab the tag length here. We'll use it to
+        // In encrypt mode, we grab the tag length here. We'll use it to
         // ensure that that allocated buffer has enough room for both the
         // final block and the auth tag. Unlike our other AES-GCM implementation
         // in CipherBase, in WebCrypto, the auth tag is concatenated to the end
         // of the generated ciphertext and returned in the same ArrayBuffer.
         tag_len = params.length;
+
+        // For OCB mode, we need to set the auth tag length
+        if (params.cipher.isOcbMode()) {
+          if (!ctx.setAeadTagLength(tag_len)) {
+            return WebCryptoCipherStatus::FAILED;
+          }
+        }
         break;
       }
       default:
@@ -108,20 +129,22 @@ WebCryptoCipherStatus AES_Cipher(Environment* env,
       .data = params.additional_data.data<unsigned char>(),
       .len = params.additional_data.size(),
   };
-  if (mode == EVP_CIPH_GCM_MODE && params.additional_data.size() &&
-      !ctx.update(buffer, nullptr, &out_len)) {
+  if ((params.cipher.isGcmMode() || params.cipher.isOcbMode()) &&
+      params.additional_data.size() && !ctx.update(buffer, nullptr, &out_len)) {
     return WebCryptoCipherStatus::FAILED;
   }
 
-  ByteSource::Builder buf(buf_len);
+  auto buf = DataPointer::Alloc(buf_len);
+  auto ptr = static_cast<unsigned char*>(buf.get());
 
   // In some outdated version of OpenSSL (e.g.
   // ubi81_sharedlibs_openssl111fips_x64) may be used in sharedlib mode, the
-  // logic will be failed when input size is zero. The newly OpenSSL has fixed
+  // logic will be failed when input size is zero. The newer OpenSSL has fixed
   // it up. But we still have to regard zero as special in Node.js code to
   // prevent old OpenSSL failure.
   //
-  // Refs: https://github.com/openssl/openssl/commit/420cb707b880e4fb649094241371701013eeb15f
+  // Refs:
+  // https://github.com/openssl/openssl/commit/420cb707b880e4fb649094241371701013eeb15f
   // Refs: https://github.com/nodejs/node/pull/38913#issuecomment-866505244
   buffer = {
       .data = in.data<unsigned char>(),
@@ -129,29 +152,35 @@ WebCryptoCipherStatus AES_Cipher(Environment* env,
   };
   if (in.empty()) {
     out_len = 0;
-  } else if (!ctx.update(buffer, buf.data<unsigned char>(), &out_len)) {
+  } else if (!ctx.update(buffer, ptr, &out_len)) {
     return WebCryptoCipherStatus::FAILED;
   }
 
   total += out_len;
   CHECK_LE(out_len, buf_len);
   out_len = ctx.getBlockSize();
-  if (!ctx.update({}, buf.data<unsigned char>() + total, &out_len, true)) {
+  if (!ctx.update({}, ptr + total, &out_len, true)) {
     return WebCryptoCipherStatus::FAILED;
   }
   total += out_len;
 
-  // If using AES_GCM, grab the generated auth tag and append
+  // If using AES_GCM or AES_OCB, grab the generated auth tag and append
   // it to the end of the ciphertext.
-  if (cipher_mode == kWebCryptoCipherEncrypt && mode == EVP_CIPH_GCM_MODE) {
-    if (!ctx.getAeadTag(tag_len, buf.data<unsigned char>() + total)) {
+  if (encrypt && (params.cipher.isGcmMode() || params.cipher.isOcbMode())) {
+    if (!ctx.getAeadTag(tag_len, ptr + total)) {
       return WebCryptoCipherStatus::FAILED;
     }
     total += tag_len;
   }
 
+  if (total == 0) {
+    *out = ByteSource::Allocated(nullptr, 0);
+    return WebCryptoCipherStatus::OK;
+  }
+
   // It's possible that we haven't used the full allocated space. Size down.
-  *out = std::move(buf).release(total);
+  buf = buf.resize(total);
+  *out = ByteSource::Allocated(buf.release());
 
   return WebCryptoCipherStatus::OK;
 }
@@ -213,11 +242,10 @@ WebCryptoCipherStatus AES_CTR_Cipher2(const KeyObjectData& key_data,
   if (!ctx) {
     return WebCryptoCipherStatus::FAILED;
   }
-  const bool encrypt = cipher_mode == kWebCryptoCipherEncrypt;
 
   if (!ctx.init(
           params.cipher,
-          encrypt,
+          cipher_mode == kWebCryptoCipherEncrypt,
           reinterpret_cast<const unsigned char*>(key_data.GetSymmetricKey()),
           counter)) {
     // Cipher init failed
@@ -230,19 +258,14 @@ WebCryptoCipherStatus AES_CTR_Cipher2(const KeyObjectData& key_data,
       .data = in.data<unsigned char>(),
       .len = in.size(),
   };
-  if (!ctx.update(buffer, out, &out_len)) {
+  if (!ctx.update(buffer, out, &out_len) ||
+      !ctx.update({}, out + out_len, &final_len, true)) {
     return WebCryptoCipherStatus::FAILED;
   }
 
-  if (!ctx.update({}, out + out_len, &final_len, true)) {
-    return WebCryptoCipherStatus::FAILED;
-  }
-
-  out_len += final_len;
-  if (static_cast<unsigned>(out_len) != in.size())
-    return WebCryptoCipherStatus::FAILED;
-
-  return WebCryptoCipherStatus::OK;
+  return static_cast<unsigned>(out_len + final_len) != in.size()
+             ? WebCryptoCipherStatus::FAILED
+             : WebCryptoCipherStatus::OK;
 }
 
 WebCryptoCipherStatus AES_CTR_Cipher(Environment* env,
@@ -258,8 +281,9 @@ WebCryptoCipherStatus AES_CTR_Cipher(Environment* env,
 
   auto num_output = BignumPointer::New();
 
-  if (!num_output.setWord(CeilDiv(in.size(), kAesBlockSize)))
+  if (!num_output.setWord(CeilDiv(in.size(), kAesBlockSize))) {
     return WebCryptoCipherStatus::FAILED;
+  }
 
   // Just like in chromium's implementation, if the counter will
   // be incremented more than there are counter values, we fail.
@@ -272,7 +296,7 @@ WebCryptoCipherStatus AES_CTR_Cipher(Environment* env,
   }
 
   // Output size is identical to the input size.
-  ByteSource::Builder buf(in.size());
+  auto buf = DataPointer::Alloc(in.size());
 
   // Also just like in chromium's implementation, if we can process
   // the input without wrapping the counter, we'll do it as a single
@@ -283,8 +307,10 @@ WebCryptoCipherStatus AES_CTR_Cipher(Environment* env,
                                   params,
                                   in,
                                   params.iv.data<unsigned char>(),
-                                  buf.data<unsigned char>());
-    if (status == WebCryptoCipherStatus::OK) *out = std::move(buf).release();
+                                  static_cast<unsigned char*>(buf.get()));
+    if (status == WebCryptoCipherStatus::OK) {
+      *out = ByteSource::Allocated(buf.release());
+    }
     return status;
   }
 
@@ -297,14 +323,16 @@ WebCryptoCipherStatus AES_CTR_Cipher(Environment* env,
                       params,
                       ByteSource::Foreign(in.data<char>(), input_size_part1),
                       params.iv.data<unsigned char>(),
-                      buf.data<unsigned char>());
+                      static_cast<unsigned char*>(buf.get()));
 
-  if (status != WebCryptoCipherStatus::OK)
+  if (status != WebCryptoCipherStatus::OK) {
     return status;
+  }
 
   // Wrap the counter around to zero
   std::vector<unsigned char> new_counter_block = BlockWithZeroedCounter(params);
 
+  auto ptr = static_cast<unsigned char*>(buf.get()) + input_size_part1;
   // Encrypt the second part...
   status =
       AES_CTR_Cipher2(key_data,
@@ -313,9 +341,11 @@ WebCryptoCipherStatus AES_CTR_Cipher(Environment* env,
                       ByteSource::Foreign(in.data<char>() + input_size_part1,
                                           in.size() - input_size_part1),
                       new_counter_block.data(),
-                      buf.data<unsigned char>() + input_size_part1);
+                      ptr);
 
-  if (status == WebCryptoCipherStatus::OK) *out = std::move(buf).release();
+  if (status == WebCryptoCipherStatus::OK) {
+    *out = ByteSource::Allocated(buf.release());
+  }
 
   return status;
 }
@@ -454,10 +484,9 @@ Maybe<void> AESCipherTraits::AdditionalConfig(
   params->variant =
       static_cast<AESKeyVariant>(args[offset].As<Uint32>()->Value());
 
-  int cipher_nid;
 #define V(name, _, nid)                                                        \
-  case kKeyVariantAES_##name: {                                                \
-    cipher_nid = nid;                                                          \
+  case AESKeyVariant::name: {                                                  \
+    params->cipher = nid;                                                      \
     break;                                                                     \
   }
   switch (params->variant) {
@@ -467,22 +496,20 @@ Maybe<void> AESCipherTraits::AdditionalConfig(
   }
 #undef V
 
-  params->cipher = Cipher::FromNid(cipher_nid);
   if (!params->cipher) {
     THROW_ERR_CRYPTO_UNKNOWN_CIPHER(env);
     return Nothing<void>();
   }
 
-  int cipher_op_mode = params->cipher.getMode();
-  if (cipher_op_mode != EVP_CIPH_WRAP_MODE) {
+  if (!params->cipher.isWrapMode()) {
     if (!ValidateIV(env, mode, args[offset + 1], params)) {
       return Nothing<void>();
     }
-    if (cipher_op_mode == EVP_CIPH_CTR_MODE) {
+    if (params->cipher.isCtrMode()) {
       if (!ValidateCounter(env, args[offset + 2], params)) {
         return Nothing<void>();
       }
-    } else if (cipher_op_mode == EVP_CIPH_GCM_MODE) {
+    } else if (params->cipher.isGcmMode() || params->cipher.isOcbMode()) {
       if (!ValidateAuthTag(env, mode, cipher_mode, args[offset + 2], params) ||
           !ValidateAdditionalData(env, mode, args[offset + 3], params)) {
         return Nothing<void>();
@@ -492,9 +519,18 @@ Maybe<void> AESCipherTraits::AdditionalConfig(
     UseDefaultIV(params);
   }
 
-  if (params->iv.size() < static_cast<size_t>(params->cipher.getIvLength())) {
-    THROW_ERR_CRYPTO_INVALID_IV(env);
-    return Nothing<void>();
+  // For OCB mode, allow variable IV lengths (1-15 bytes)
+  if (params->cipher.isOcbMode()) {
+    if (params->iv.size() == 0 || params->iv.size() > 15) {
+      THROW_ERR_CRYPTO_INVALID_IV(env);
+      return Nothing<void>();
+    }
+  } else {
+    // For other modes, check against the cipher's expected IV length
+    if (params->iv.size() < static_cast<size_t>(params->cipher.getIvLength())) {
+      THROW_ERR_CRYPTO_INVALID_IV(env);
+      return Nothing<void>();
+    }
   }
 
   return JustVoid();
@@ -507,7 +543,7 @@ WebCryptoCipherStatus AESCipherTraits::DoCipher(Environment* env,
                                                 const ByteSource& in,
                                                 ByteSource* out) {
 #define V(name, fn, _)                                                         \
-  case kKeyVariantAES_##name:                                                  \
+  case AESKeyVariant::name:                                                    \
     return fn(env, key_data, cipher_mode, params, in, out);
   switch (params.variant) {
     VARIANTS(V)
@@ -520,8 +556,13 @@ WebCryptoCipherStatus AESCipherTraits::DoCipher(Environment* env,
 void AES::Initialize(Environment* env, Local<Object> target) {
   AESCryptoJob::Initialize(env, target);
 
-#define V(name, _, __) NODE_DEFINE_CONSTANT(target, kKeyVariantAES_##name);
+#define V(name, _, __)                                                         \
+  constexpr static auto kKeyVariantAES_##name =                                \
+      static_cast<int>(AESKeyVariant::name);                                   \
+  NODE_DEFINE_CONSTANT(target, kKeyVariantAES_##name);
+
   VARIANTS(V)
+
 #undef V
 }
 
