@@ -1,6 +1,7 @@
 #include "node_modules.h"
 #include <cstdio>
 #include "base_object-inl.h"
+#include "compile_cache.h"
 #include "node_errors.h"
 #include "node_external_reference.h"
 #include "node_url.h"
@@ -21,12 +22,16 @@ namespace modules {
 
 using v8::Array;
 using v8::Context;
+using v8::External;
 using v8::FunctionCallbackInfo;
 using v8::HandleScope;
+using v8::Integer;
 using v8::Isolate;
 using v8::Local;
 using v8::LocalVector;
+using v8::Name;
 using v8::NewStringType;
+using v8::Null;
 using v8::Object;
 using v8::ObjectTemplate;
 using v8::Primitive;
@@ -102,23 +107,11 @@ const BindingData::PackageConfig* BindingData::GetPackageJSON(
   if (ReadFileSync(&package_config.raw_json, path.data()) < 0) {
     return nullptr;
   }
-  // In some systems, std::string is annotated to generate an
-  // AddressSanitizer: container-overflow error when reading beyond the end of
-  // the string even when we are still within the capacity of the string.
-  // https://github.com/google/sanitizers/wiki/AddressSanitizerContainerOverflow
-  // https://github.com/nodejs/node/issues/55584
-  // The next lines are a workaround to avoid this false positive.
-  size_t json_length = package_config.raw_json.size();
-  package_config.raw_json.append(simdjson::SIMDJSON_PADDING, ' ');
-  simdjson::padded_string_view json_view(package_config.raw_json.data(),
-                                         json_length,
-                                         package_config.raw_json.size());
-  // End of workaround
-
   simdjson::ondemand::document document;
   simdjson::ondemand::object main_object;
   simdjson::error_code error =
-      binding_data->json_parser.iterate(json_view).get(document);
+      binding_data->json_parser.iterate(simdjson::pad(package_config.raw_json))
+          .get(document);
 
   const auto throw_invalid_package_config = [error_context, path, realm]() {
     if (error_context == nullptr) {
@@ -327,40 +320,6 @@ const BindingData::PackageConfig* BindingData::TraverseParent(
   return nullptr;
 }
 
-void BindingData::GetNearestParentPackageJSON(
-    const v8::FunctionCallbackInfo<v8::Value>& args) {
-  CHECK_GE(args.Length(), 1);
-  CHECK(args[0]->IsString());
-
-  Realm* realm = Realm::GetCurrent(args);
-  BufferValue path_value(realm->isolate(), args[0]);
-  // Check if the path has a trailing slash. If so, add it after
-  // ToNamespacedPath() as it will be deleted by ToNamespacedPath()
-  bool slashCheck = path_value.ToStringView().ends_with(kPathSeparator);
-
-  ToNamespacedPath(realm->env(), &path_value);
-
-  std::string path_value_str = path_value.ToString();
-  if (slashCheck) {
-    path_value_str.push_back(kPathSeparator);
-  }
-
-  std::filesystem::path path;
-
-#ifdef _WIN32
-  std::wstring wide_path = ConvertToWideString(path_value_str, GetACP());
-  path = std::filesystem::path(wide_path);
-#else
-  path = std::filesystem::path(path_value_str);
-#endif
-
-  auto package_json = TraverseParent(realm, path);
-
-  if (package_json != nullptr) {
-    args.GetReturnValue().Set(package_json->Serialize(realm));
-  }
-}
-
 void BindingData::GetNearestParentPackageJSONType(
     const FunctionCallbackInfo<Value>& args) {
   CHECK_GE(args.Length(), 1);
@@ -394,9 +353,10 @@ void BindingData::GetNearestParentPackageJSONType(
     return;
   }
 
-  Local<Value> value =
-      ToV8Value(realm->context(), package_json->type).ToLocalChecked();
-  args.GetReturnValue().Set(value);
+  Local<Value> value;
+  if (ToV8Value(realm->context(), package_json->type).ToLocal(&value)) {
+    args.GetReturnValue().Set(value);
+  }
 }
 
 template <bool return_only_type>
@@ -473,12 +433,11 @@ void BindingData::GetPackageScopeConfig(
   auto package_json_url_as_path =
       url::FileURLToPath(realm->env(), *package_json_url);
   CHECK(package_json_url_as_path);
-  return args.GetReturnValue().Set(
-      String::NewFromUtf8(realm->isolate(),
-                          package_json_url_as_path->c_str(),
-                          NewStringType::kNormal,
-                          package_json_url_as_path->size())
-          .ToLocalChecked());
+  Local<Value> ret;
+  if (ToV8Value(realm->context(), *package_json_url_as_path, realm->isolate())
+          .ToLocal(&ret)) {
+    args.GetReturnValue().Set(ret);
+  }
 }
 
 void FlushCompileCache(const FunctionCallbackInfo<Value>& args) {
@@ -510,11 +469,13 @@ void EnableCompileCache(const FunctionCallbackInfo<Value>& args) {
   }
   Utf8Value value(isolate, args[0]);
   CompileCacheEnableResult result = env->EnableCompileCache(*value);
-  Local<Value> values[] = {
-      v8::Integer::New(isolate, static_cast<uint8_t>(result.status)),
-      ToV8Value(context, result.message).ToLocalChecked(),
-      ToV8Value(context, result.cache_directory).ToLocalChecked()};
-  args.GetReturnValue().Set(Array::New(isolate, &values[0], arraysize(values)));
+  Local<Value> values[3];
+  values[0] = v8::Integer::New(isolate, static_cast<uint8_t>(result.status));
+  if (ToV8Value(context, result.message).ToLocal(&values[1]) &&
+      ToV8Value(context, result.cache_directory).ToLocal(&values[2])) {
+    args.GetReturnValue().Set(
+        Array::New(isolate, &values[0], arraysize(values)));
+  }
 }
 
 void GetCompileCacheDir(const FunctionCallbackInfo<Value>& args) {
@@ -525,9 +486,67 @@ void GetCompileCacheDir(const FunctionCallbackInfo<Value>& args) {
     args.GetReturnValue().Set(v8::String::Empty(isolate));
     return;
   }
-  args.GetReturnValue().Set(
-      ToV8Value(context, env->compile_cache_handler()->cache_dir())
-          .ToLocalChecked());
+
+  Local<Value> ret;
+  if (ToV8Value(context, env->compile_cache_handler()->cache_dir())
+          .ToLocal(&ret)) {
+    args.GetReturnValue().Set(ret);
+  }
+}
+
+void GetCompileCacheEntry(const FunctionCallbackInfo<Value>& args) {
+  Isolate* isolate = args.GetIsolate();
+  CHECK(args[0]->IsString());  // TODO(joyeecheung): accept buffer.
+  CHECK(args[1]->IsString());
+  CHECK(args[2]->IsUint32());
+  Local<Context> context = isolate->GetCurrentContext();
+  Environment* env = Environment::GetCurrent(context);
+  if (!env->use_compile_cache()) {
+    return;
+  }
+  Local<String> source = args[0].As<String>();
+  Local<String> filename = args[1].As<String>();
+  CachedCodeType type =
+      static_cast<CachedCodeType>(args[2].As<v8::Uint32>()->Value());
+  auto* cache_entry =
+      env->compile_cache_handler()->GetOrInsert(source, filename, type);
+  if (cache_entry == nullptr) {
+    return;
+  }
+
+  v8::LocalVector<v8::Name> names(isolate,
+                                  {FIXED_ONE_BYTE_STRING(isolate, "external")});
+  v8::LocalVector<v8::Value> values(isolate,
+                                    {v8::External::New(isolate, cache_entry)});
+  if (cache_entry->cache != nullptr) {
+    Debug(env,
+          DebugCategory::COMPILE_CACHE,
+          "[compile cache] retrieving transpile cache for %s %s...",
+          cache_entry->type_name(),
+          cache_entry->source_filename);
+
+    std::string_view cache(
+        reinterpret_cast<const char*>(cache_entry->cache->data),
+        cache_entry->cache->length);
+    Local<Value> transpiled;
+    // TODO(joyeecheung): convert with simdutf and into external strings
+    if (!ToV8Value(context, cache).ToLocal(&transpiled)) {
+      Debug(env, DebugCategory::COMPILE_CACHE, "failed\n");
+      return;
+    } else {
+      Debug(env, DebugCategory::COMPILE_CACHE, "success\n");
+    }
+    names.push_back(FIXED_ONE_BYTE_STRING(isolate, "transpiled"));
+    values.push_back(transpiled);
+  } else {
+    Debug(env,
+          DebugCategory::COMPILE_CACHE,
+          "[compile cache] no transpile cache for %s %s\n",
+          cache_entry->type_name(),
+          cache_entry->source_filename);
+  }
+  args.GetReturnValue().Set(Object::New(
+      isolate, v8::Null(isolate), names.data(), values.data(), names.size()));
 }
 
 static void PathHelpersLazyGetter(Local<v8::Name> name,
@@ -600,6 +619,18 @@ void InitImportMetaPathHelpers(const FunctionCallbackInfo<Value>& args) {
           .IsNothing())
     return;
 }
+void SaveCompileCacheEntry(const FunctionCallbackInfo<Value>& args) {
+  Isolate* isolate = args.GetIsolate();
+  Local<Context> context = isolate->GetCurrentContext();
+  Environment* env = Environment::GetCurrent(context);
+  DCHECK(env->use_compile_cache());
+  CHECK(args[0]->IsExternal());
+  CHECK(args[1]->IsString());  // TODO(joyeecheung): accept buffer.
+  auto* cache_entry =
+      static_cast<CompileCacheEntry*>(args[0].As<External>()->Value());
+  Utf8Value utf8(isolate, args[1].As<String>());
+  env->compile_cache_handler()->MaybeSave(cache_entry, utf8.ToStringView());
+}
 
 void BindingData::CreatePerIsolateProperties(IsolateData* isolate_data,
                                              Local<ObjectTemplate> target) {
@@ -609,16 +640,14 @@ void BindingData::CreatePerIsolateProperties(IsolateData* isolate_data,
             target,
             "getNearestParentPackageJSONType",
             GetNearestParentPackageJSONType);
-  SetMethod(isolate,
-            target,
-            "getNearestParentPackageJSON",
-            GetNearestParentPackageJSON);
   SetMethod(
       isolate, target, "getPackageScopeConfig", GetPackageScopeConfig<false>);
   SetMethod(isolate, target, "getPackageType", GetPackageScopeConfig<true>);
   SetMethod(isolate, target, "enableCompileCache", EnableCompileCache);
   SetMethod(isolate, target, "getCompileCacheDir", GetCompileCacheDir);
   SetMethod(isolate, target, "flushCompileCache", FlushCompileCache);
+  SetMethod(isolate, target, "getCompileCacheEntry", GetCompileCacheEntry);
+  SetMethod(isolate, target, "saveCompileCacheEntry", SaveCompileCacheEntry);
   SetMethod(isolate, target, "setLazyPathHelpers", InitImportMetaPathHelpers);
 }
 
@@ -636,24 +665,44 @@ void BindingData::CreatePerContextProperties(Local<Object> target,
   compile_cache_status_values.push_back(                                       \
       FIXED_ONE_BYTE_STRING(isolate, #status));
   COMPILE_CACHE_STATUS(V)
+#undef V
 
   USE(target->Set(context,
                   FIXED_ONE_BYTE_STRING(isolate, "compileCacheStatus"),
                   Array::New(isolate,
                              compile_cache_status_values.data(),
                              compile_cache_status_values.size())));
+
+  LocalVector<Name> cached_code_type_keys(isolate);
+  LocalVector<Value> cached_code_type_values(isolate);
+
+#define V(type, value)                                                         \
+  cached_code_type_keys.push_back(FIXED_ONE_BYTE_STRING(isolate, #type));      \
+  cached_code_type_values.push_back(Integer::New(isolate, value));             \
+  DCHECK_EQ(value, cached_code_type_values.size() - 1);
+  CACHED_CODE_TYPES(V)
+#undef V
+
+  USE(target->Set(context,
+                  FIXED_ONE_BYTE_STRING(isolate, "cachedCodeTypes"),
+                  Object::New(isolate,
+                              Null(isolate),
+                              cached_code_type_keys.data(),
+                              cached_code_type_values.data(),
+                              cached_code_type_keys.size())));
 }
 
 void BindingData::RegisterExternalReferences(
     ExternalReferenceRegistry* registry) {
   registry->Register(ReadPackageJSON);
   registry->Register(GetNearestParentPackageJSONType);
-  registry->Register(GetNearestParentPackageJSON);
   registry->Register(GetPackageScopeConfig<false>);
   registry->Register(GetPackageScopeConfig<true>);
   registry->Register(EnableCompileCache);
   registry->Register(GetCompileCacheDir);
   registry->Register(FlushCompileCache);
+  registry->Register(GetCompileCacheEntry);
+  registry->Register(SaveCompileCacheEntry);
   registry->Register(InitImportMetaPathHelpers);
 }
 
